class QueueItem:
    def __init__(self, id=None, bitrate=None, title=None, artist=None, cover=None, explicit=False, size=None, type=None, settings=None, queueItemDict=None):
        if queueItemDict:
            self.title = queueItemDict['title']
            self.artist = queueItemDict['artist']
            self.cover = queueItemDict['cover']
            self.explicit = queueItemDict.get('explicit', False)
            self.size = queueItemDict['size']
            self.type = queueItemDict['type']
            self.id = queueItemDict['id']
            self.bitrate = queueItemDict['bitrate']
            self.extrasPath = queueItemDict.get('extrasPath', '')
            self.files = queueItemDict['files']
            self.downloaded = queueItemDict['downloaded']
            self.failed = queueItemDict['failed']
            self.errors = queueItemDict['errors']
            self.progress = queueItemDict['progress']
            self.settings = queueItemDict.get('settings')
        else:
            self.title = title
            self.artist = artist
            self.cover = cover
            self.explicit = explicit
            self.size = size
            self.type = type
            self.id = id
            self.bitrate = bitrate
            self.extrasPath = None
            self.files = []
            self.settings = settings
            self.downloaded = 0
            self.failed = 0
            self.errors = []
            self.progress = 0
        self.uuid = f"{self.type}_{self.id}_{self.bitrate}"
        self.cancel = False
        self.ack = None

    def toDict(self):
        return {
            'title': self.title,
            'artist': self.artist,
            'cover': self.cover,
            'explicit': self.explicit,
            'size': self.size,
            'extrasPath': self.extrasPath,
            'files': self.files,
            'downloaded': self.downloaded,
            'failed': self.failed,
            'errors': self.errors,
            'progress': self.progress,
            'type': self.type,
            'id': self.id,
            'bitrate': self.bitrate,
            'uuid': self.uuid,
            'ack': self.ack
        }

    def getResettedItem(self):
        item = self.toDict()
        item['downloaded'] = 0
        item['failed'] = 0
        item['progress'] = 0
        item['errors'] = []
        return item

    def getSlimmedItem(self):
        light = self.toDict()
        propertiesToDelete = ['single', 'collection', '_EXTRA', 'settings']
        for property in propertiesToDelete:
            if property in light:
                del light[property]
        return light

class QISingle(QueueItem):
    def __init__(self, id=None, bitrate=None, title=None, artist=None, cover=None, explicit=False, type=None, settings=None, single=None, queueItemDict=None):
        if queueItemDict:
            super().__init__(queueItemDict=queueItemDict)
            self.single = queueItemDict['single']
        else:
            super().__init__(id, bitrate, title, artist, cover, explicit, 1, type, settings)
            self.single = single

    def toDict(self):
        queueItem = super().toDict()
        queueItem['single'] = self.single
        return queueItem

class QICollection(QueueItem):
    def __init__(self, id=None, bitrate=None, title=None, artist=None, cover=None, explicit=False, size=None, type=None, settings=None, collection=None, queueItemDict=None):
        if queueItemDict:
            super().__init__(queueItemDict=queueItemDict)
            self.collection = queueItemDict['collection']
        else:
            super().__init__(id, bitrate, title, artist, cover, explicit, size, type, settings)
            self.collection = collection

    def toDict(self):
        queueItem = super().toDict()
        queueItem['collection'] = self.collection
        return queueItem

class QIConvertable(QICollection):
    def __init__(self, id=None, bitrate=None, title=None, artist=None, cover=None, explicit=False, size=None, type=None, settings=None, extra=None, queueItemDict=None):
        if queueItemDict:
            super().__init__(queueItemDict=queueItemDict)
            self.extra = queueItemDict['_EXTRA']
        else:
            super().__init__(id, bitrate, title, artist, cover, explicit, size, type, settings, [])
            self.extra = extra

    def toDict(self):
        queueItem = super().toDict()
        queueItem['_EXTRA'] = self.extra
        return queueItem



from pathlib import Path
import sys
from os import getenv

userdata = ""
homedata = Path.home()

if getenv("APPDATA"):
    userdata = Path(getenv("APPDATA")) / "deemix"
elif sys.platform.startswith('darwin'):
    userdata = homedata / 'Library' / 'Application Support' / 'deemix'
elif getenv("XDG_CONFIG_HOME"):
    userdata = Path(getenv("XDG_CONFIG_HOME")) / 'deemix'
else:
    userdata = homedata / '.config' / 'deemix'

def getHomeFolder():
    return homedata

def getConfigFolder():
    return userdata



from pathlib import Path
from os import makedirs

from deemix.app import deemix

class cli(deemix):
    def __init__(self, downloadpath, configFolder=None):
        super().__init__(configFolder)
        if downloadpath:
            self.set.settings['downloadLocation'] = str(downloadpath)
            makedirs(downloadpath, exist_ok=True)
            print("Using folder: "+self.set.settings['downloadLocation'])

    def downloadLink(self, url, bitrate=None):
        for link in url:
            if ';' in link:
                for l in link.split(";"):
                    self.qm.addToQueue(self.dz, l, self.set.settings, bitrate)
            else:
                self.qm.addToQueue(self.dz, link, self.set.settings, bitrate)

    def requestValidArl(self):
        while True:
            arl = input("Paste here your arl:")
            if self.dz.login_via_arl(arl):
                break
        return arl

    def login(self):
        configFolder = Path(self.set.configFolder)
        if not configFolder.is_dir():
            makedirs(configFolder, exist_ok=True)
        if (configFolder / '.arl').is_file():
            with open(configFolder / '.arl', 'r') as f:
                arl = f.readline().rstrip("\n")
            if not self.dz.login_via_arl(arl):
                arl = self.requestValidArl()
        else:
            arl = self.requestValidArl()
        with open(configFolder / '.arl', 'w') as f:
            f.write(arl)



import json
from pathlib import Path
from os import makedirs, listdir
from deemix import __version__ as deemixVersion
from deemix.api.deezer import TrackFormats
import logging
import datetime
import platform

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

import deemix.utils.localpaths as localpaths

class OverwriteOption():
    """Should the lib overwrite files?"""

    OVERWRITE = 'y'
    """Yes, overwrite the file"""

    DONT_OVERWRITE = 'n'
    """No, don't overwrite the file"""

    DONT_CHECK_EXT = 'e'
    """No, and don't check for extensions"""

    KEEP_BOTH = 'b'
    """No, and keep both files"""

    ONLY_TAGS = 't'
    """Overwrite only the tags"""

class FeaturesOption():
    """What should I do with featured artists?"""

    NO_CHANGE = 0
    """Do nothing"""

    REMOVE_TITLE = 1
    """Remove from track title"""

    REMOVE_TITLE_ALBUM = 3
    """Remove from track title and album title"""

    MOVE_TITLE = 2
    """Move to track title"""

DEFAULT_SETTINGS = {
  "downloadLocation": str(localpaths.getHomeFolder() / 'deemix Music'),
  "tracknameTemplate": "%artist% - %title%",
  "albumTracknameTemplate": "%tracknumber% - %title%",
  "playlistTracknameTemplate": "%position% - %artist% - %title%",
  "createPlaylistFolder": True,
  "playlistNameTemplate": "%playlist%",
  "createArtistFolder": False,
  "artistNameTemplate": "%artist%",
  "createAlbumFolder": True,
  "albumNameTemplate": "%artist% - %album%",
  "createCDFolder": True,
  "createStructurePlaylist": False,
  "createSingleFolder": False,
  "padTracks": True,
  "paddingSize": "0",
  "illegalCharacterReplacer": "_",
  "queueConcurrency": 3,
  "maxBitrate": str(TrackFormats.MP3_320),
  "fallbackBitrate": True,
  "fallbackSearch": False,
  "logErrors": True,
  "logSearched": False,
  "saveDownloadQueue": False,
  "overwriteFile": OverwriteOption.DONT_OVERWRITE,
  "createM3U8File": False,
  "playlistFilenameTemplate": "playlist",
  "syncedLyrics": False,
  "embeddedArtworkSize": 800,
  "embeddedArtworkPNG": False,
  "localArtworkSize": 1400,
  "localArtworkFormat": "jpg",
  "saveArtwork": True,
  "coverImageTemplate": "cover",
  "saveArtworkArtist": False,
  "artistImageTemplate": "folder",
  "jpegImageQuality": 80,
  "dateFormat": "Y-M-D",
  "albumVariousArtists": True,
  "removeAlbumVersion": False,
  "removeDuplicateArtists": False,
  "tagsLanguage": "",
  "featuredToTitle": FeaturesOption.NO_CHANGE,
  "titleCasing": "nothing",
  "artistCasing": "nothing",
  "executeCommand": "",
  "tags": {
    "title": True,
    "artist": True,
    "album": True,
    "cover": True,
    "trackNumber": True,
    "trackTotal": False,
    "discNumber": True,
    "discTotal": False,
    "albumArtist": True,
    "genre": True,
    "year": True,
    "date": True,
    "explicit": False,
    "isrc": True,
    "length": True,
    "barcode": True,
    "bpm": True,
    "replayGain": False,
    "label": True,
    "lyrics": False,
    "syncedLyrics": False,
    "copyright": False,
    "composer": False,
    "involvedPeople": False,
    "savePlaylistAsCompilation": False,
    "useNullSeparator": False,
    "saveID3v1": True,
    "multiArtistSeparator": "default",
    "singleAlbumArtist": False,
    "coverDescriptionUTF8": False
  }
}

class Settings:
    def __init__(self, configFolder=None):
        self.settings = {}
        self.configFolder = Path(configFolder or localpaths.getConfigFolder())

        # Create config folder if it doesn't exsist
        makedirs(self.configFolder, exist_ok=True)

        # Create config file if it doesn't exsist
        if not (self.configFolder / 'config.json').is_file():
            with open(self.configFolder / 'config.json', 'w') as f:
                json.dump(DEFAULT_SETTINGS, f, indent=2)

        # Read config file
        with open(self.configFolder / 'config.json', 'r') as configFile:
            self.settings = json.load(configFile)

        self.settingsCheck()

        # Make sure the download path exsits
        makedirs(self.settings['downloadLocation'], exist_ok=True)

        # LOGFILES

        # Create logfile name and path
        logspath = self.configFolder / 'logs'
        now = datetime.datetime.now()
        logfile = now.strftime("%Y-%m-%d_%H%M%S")+".log"
        makedirs(logspath, exist_ok=True)

        # Add handler for logging
        fh = logging.FileHandler(logspath / logfile, 'w', 'utf-8')
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(logging.Formatter('%(asctime)s - [%(levelname)s] %(message)s'))
        logger.addHandler(fh)
        logger.info(f"{platform.platform(True, True)} - Python {platform.python_version()}, deemix {deemixVersion}")

        # Only keep last 5 logfiles (to preserve disk space)
        logslist = listdir(logspath)
        logslist.sort()
        if len(logslist)>5:
            for i in range(len(logslist)-5):
                (logspath / logslist[i]).unlink()

    # Saves the settings
    def saveSettings(self, newSettings=None, dz=None):
        if newSettings:
            if dz and newSettings.get('tagsLanguage') != self.settings.get('tagsLanguage'): dz.set_accept_language(newSettings.get('tagsLanguage'))
            self.settings = newSettings
        with open(self.configFolder / 'config.json', 'w') as configFile:
            json.dump(self.settings, configFile, indent=2)

    # Checks if the default settings have changed
    def settingsCheck(self):
        changes = 0
        for set in DEFAULT_SETTINGS:
            if not set in self.settings or type(self.settings[set]) != type(DEFAULT_SETTINGS[set]):
                self.settings[set] = DEFAULT_SETTINGS[set]
                changes += 1
        for set in DEFAULT_SETTINGS['tags']:
            if not set in self.settings['tags'] or type(self.settings['tags'][set]) != type(DEFAULT_SETTINGS['tags'][set]):
                self.settings['tags'][set] = DEFAULT_SETTINGS['tags'][set]
                changes += 1
        if self.settings['downloadLocation'] == "":
            self.settings['downloadLocation'] = str(localpaths.getHomeFolder() / 'deemix Music')
            changes += 1
        for template in ['tracknameTemplate', 'albumTracknameTemplate', 'playlistTracknameTemplate', 'playlistNameTemplate', 'artistNameTemplate', 'albumNameTemplate', 'playlistFilenameTemplate', 'coverImageTemplate', 'artistImageTemplate', 'paddingSize']:
            if self.settings[template] == "":
                self.settings[template] = DEFAULT_SETTINGS[template]
                changes += 1
        if changes > 0:
            self.saveSettings()



import re
from os.path import sep as pathSep
from pathlib import Path
from unicodedata import normalize
from deemix.api.deezer import TrackFormats

bitrateLabels = {
    TrackFormats.MP4_RA3: "360 HQ",
    TrackFormats.MP4_RA2: "360 MQ",
    TrackFormats.MP4_RA1: "360 LQ",
    TrackFormats.FLAC   : "FLAC",
    TrackFormats.MP3_320: "320",
    TrackFormats.MP3_128: "128",
    TrackFormats.DEFAULT: "128",
    TrackFormats.LOCAL  : "MP3"
}

def fixName(txt, char='_'):
    txt = str(txt)
    txt = re.sub(r'[\0\/\\:*?"<>|]', char, txt)
    txt = normalize("NFC", txt)
    return txt

def fixEndOfData(bString):
    try:
        bString.decode()
        return True
    except:
        return False

def fixLongName(name):
    if pathSep in name:
        sepName = name.split(pathSep)
        name = ""
        for txt in sepName:
            txt = fixLongName(txt)
            name += txt + pathSep
        name = name[:-1]
    else:
        name = name.encode('utf-8')[:200]
        while not fixEndOfData(name):
            name = name[:-1]
        name = name.decode()
    return name


def antiDot(string):
    while string[-1:] == "." or string[-1:] == " " or string[-1:] == "\n":
        string = string[:-1]
    if len(string) < 1:
        string = "dot"
    return string


def pad(num, max, settings):
    if int(settings['paddingSize']) == 0:
        paddingSize = len(str(max))
    else:
        paddingSize = len(str(10 ** (int(settings['paddingSize']) - 1)))
    if paddingSize == 1:
        paddingSize = 2
    if settings['padTracks']:
        return str(num).zfill(paddingSize)
    else:
        return str(num)

def generateFilename(track, settings, template):
    filename = template or "%artist% - %title%"
    return settingsRegex(filename, track, settings)

def generateFilepath(track, settings):
    filepath = Path(settings['downloadLocation'])
    artistPath = None
    coverPath = None
    extrasPath = None

    if settings['createPlaylistFolder'] and track.playlist and not settings['tags']['savePlaylistAsCompilation']:
        filepath = filepath / settingsRegexPlaylist(settings['playlistNameTemplate'], track.playlist, settings)

    if track.playlist and not settings['tags']['savePlaylistAsCompilation']:
        extrasPath = filepath

    if (
        (settings['createArtistFolder'] and not track.playlist) or
        (settings['createArtistFolder'] and track.playlist and settings['tags']['savePlaylistAsCompilation']) or
        (settings['createArtistFolder'] and track.playlist and settings['createStructurePlaylist'])
    ):
        filepath = filepath / settingsRegexArtist(settings['artistNameTemplate'], track.album['mainArtist'], settings)
        artistPath = filepath

    if (settings['createAlbumFolder'] and
            (not track.singleDownload or (track.singleDownload and settings['createSingleFolder'])) and
            (not track.playlist or
                (track.playlist and settings['tags']['savePlaylistAsCompilation']) or
                (track.playlist and settings['createStructurePlaylist'])
            )
    ):
        filepath = filepath / settingsRegexAlbum(settings['albumNameTemplate'], track.album, settings, track.playlist)
        coverPath = filepath

    if not (track.playlist and not settings['tags']['savePlaylistAsCompilation']):
        extrasPath = filepath

    if (
            int(track.album['discTotal']) > 1 and (
            (settings['createAlbumFolder'] and settings['createCDFolder']) and
            (not track.singleDownload or (track.singleDownload and settings['createSingleFolder'])) and
            (not track.playlist or
                (track.playlist and settings['tags']['savePlaylistAsCompilation']) or
                (track.playlist and settings['createStructurePlaylist'])
            )
    )):
        filepath = filepath / f'CD{str(track.discNumber)}'

    return (filepath, artistPath, coverPath, extrasPath)


def settingsRegex(filename, track, settings):
    filename = filename.replace("%title%", fixName(track.title, settings['illegalCharacterReplacer']))
    filename = filename.replace("%artist%", fixName(track.mainArtist['name'], settings['illegalCharacterReplacer']))
    filename = filename.replace("%artists%", fixName(", ".join(track.artists), settings['illegalCharacterReplacer']))
    filename = filename.replace("%allartists%", fixName(track.artistsString, settings['illegalCharacterReplacer']))
    filename = filename.replace("%mainartists%", fixName(track.mainArtistsString, settings['illegalCharacterReplacer']))
    if track.featArtistsString:
        filename = filename.replace("%featartists%", fixName('('+track.featArtistsString+')', settings['illegalCharacterReplacer']))
    else:
        filename = filename.replace("%featartists%", '')
    filename = filename.replace("%album%", fixName(track.album['title'], settings['illegalCharacterReplacer']))
    filename = filename.replace("%albumartist%", fixName(track.album['mainArtist']['name'], settings['illegalCharacterReplacer']))
    filename = filename.replace("%tracknumber%", pad(track.trackNumber, track.album['trackTotal'], settings))
    filename = filename.replace("%tracktotal%", str(track.album['trackTotal']))
    filename = filename.replace("%discnumber%", str(track.discNumber))
    filename = filename.replace("%disctotal%", str(track.album['discTotal']))
    if len(track.album['genre']) > 0:
        filename = filename.replace("%genre%",
                                    fixName(track.album['genre'][0], settings['illegalCharacterReplacer']))
    else:
        filename = filename.replace("%genre%", "Unknown")
    filename = filename.replace("%year%", str(track.date['year']))
    filename = filename.replace("%date%", track.dateString)
    filename = filename.replace("%bpm%", str(track.bpm))
    filename = filename.replace("%label%", fixName(track.album['label'], settings['illegalCharacterReplacer']))
    filename = filename.replace("%isrc%", track.ISRC)
    filename = filename.replace("%upc%", track.album['barcode'])
    filename = filename.replace("%explicit%", "(Explicit)" if track.explicit else "")

    filename = filename.replace("%track_id%", str(track.id))
    filename = filename.replace("%album_id%", str(track.album['id']))
    filename = filename.replace("%artist_id%", str(track.mainArtist['id']))
    if track.playlist:
        filename = filename.replace("%playlist_id%", str(track.playlist['playlistId']))
        filename = filename.replace("%position%", pad(track.position, track.playlist['trackTotal'], settings))
    else:
        filename = filename.replace("%playlist_id%", '')
        filename = filename.replace("%position%", pad(track.trackNumber, track.album['trackTotal'], settings))
    filename = filename.replace('\\', pathSep).replace('/', pathSep)
    return antiDot(fixLongName(filename))


def settingsRegexAlbum(foldername, album, settings, playlist=None):
    if playlist and settings['tags']['savePlaylistAsCompilation']:
        foldername = foldername.replace("%album_id%", "pl_" + str(playlist['playlistId']))
        foldername = foldername.replace("%genre%", "Compile")
    else:
        foldername = foldername.replace("%album_id%", str(album['id']))
        if len(album['genre']) > 0:
            foldername = foldername.replace("%genre%", fixName(album['genre'][0], settings['illegalCharacterReplacer']))
        else:
            foldername = foldername.replace("%genre%", "Unknown")
    foldername = foldername.replace("%album%", fixName(album['title'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%artist%",
                                    fixName(album['mainArtist']['name'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%artist_id%", str(album['mainArtist']['id']))
    foldername = foldername.replace("%tracktotal%", str(album['trackTotal']))
    foldername = foldername.replace("%disctotal%", str(album['discTotal']))
    foldername = foldername.replace("%type%", fixName(album['recordType'].capitalize(), settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%upc%", album['barcode'])
    foldername = foldername.replace("%explicit%", "(Explicit)" if album['explicit'] else "")
    foldername = foldername.replace("%label%", fixName(album['label'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%year%", str(album['date']['year']))
    foldername = foldername.replace("%date%", album['dateString'])
    foldername = foldername.replace("%bitrate%", bitrateLabels[int(album['bitrate'])])

    foldername = foldername.replace('\\', pathSep).replace('/', pathSep)
    return antiDot(fixLongName(foldername))


def settingsRegexArtist(foldername, artist, settings):
    foldername = foldername.replace("%artist%", fixName(artist['name'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%artist_id%", str(artist['id']))
    foldername = foldername.replace('\\', pathSep).replace('/', pathSep)
    return antiDot(fixLongName(foldername))


def settingsRegexPlaylist(foldername, playlist, settings):
    foldername = foldername.replace("%playlist%", fixName(playlist['title'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%playlist_id%", fixName(playlist['playlistId'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%owner%", fixName(playlist['owner']['name'], settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%owner_id%", str(playlist['owner']['id']))
    foldername = foldername.replace("%year%", str(playlist['date']['year']))
    foldername = foldername.replace("%date%", str(playlist['dateString']))
    foldername = foldername.replace("%explicit%", "(Explicit)" if playlist['explicit'] else "")
    foldername = foldername.replace('\\', pathSep).replace('/', pathSep)
    return antiDot(fixLongName(foldername))

def settingsRegexPlaylistFile(foldername, queueItem, settings):
    foldername = foldername.replace("%title%", fixName(queueItem.title, settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%artist%", fixName(queueItem.artist, settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%size%", str(queueItem.size))
    foldername = foldername.replace("%type%", fixName(queueItem.type, settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%id%", fixName(queueItem.id, settings['illegalCharacterReplacer']))
    foldername = foldername.replace("%bitrate%", bitrateLabels[int(queueItem.bitrate)])
    foldername = foldername.replace('\\', pathSep).replace('/', pathSep).replace(pathSep, settings['illegalCharacterReplacer'])
    return antiDot(fixLongName(foldername))



import eventlet
from eventlet.green.subprocess import call as execute
requests = eventlet.import_patched('requests')
get = requests.get
request_exception = requests.exceptions

from os.path import sep as pathSep
from pathlib import Path
from shlex import quote
import re
import errno

from ssl import SSLError
from os import makedirs
from tempfile import gettempdir

from deemix.app.queueitem import QISingle, QICollection
from deemix.app.track import Track, AlbumDoesntExsists
from deemix.utils import changeCase
from deemix.utils.pathtemplates import generateFilename, generateFilepath, settingsRegexAlbum, settingsRegexArtist, settingsRegexPlaylistFile
from deemix.api.deezer import USER_AGENT_HEADER, TrackFormats
from deemix.utils.taggers import tagID3, tagFLAC
from deemix.app.settings import OverwriteOption, FeaturesOption

from Cryptodome.Cipher import Blowfish
from mutagen.flac import FLACNoHeaderError, error as FLACError
import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

TEMPDIR = Path(gettempdir()) / 'deemix-imgs'
if not TEMPDIR.is_dir(): makedirs(TEMPDIR)

extensions = {
    TrackFormats.FLAC:    '.flac',
    TrackFormats.LOCAL:   '.mp3',
    TrackFormats.MP3_320: '.mp3',
    TrackFormats.MP3_128: '.mp3',
    TrackFormats.DEFAULT: '.mp3',
    TrackFormats.MP4_RA3: '.mp4',
    TrackFormats.MP4_RA2: '.mp4',
    TrackFormats.MP4_RA1: '.mp4'
}

errorMessages = {
    'notOnDeezer': "Track not available on Deezer!",
    'notEncoded': "Track not yet encoded!",
    'notEncodedNoAlternative': "Track not yet encoded and no alternative found!",
    'wrongBitrate': "Track not found at desired bitrate.",
    'wrongBitrateNoAlternative': "Track not found at desired bitrate and no alternative found!",
    'no360RA': "Track is not available in Reality Audio 360.",
    'notAvailable': "Track not available on deezer's servers!",
    'notAvailableNoAlternative': "Track not available on deezer's servers and no alternative found!",
    'noSpaceLeft': "No space left on target drive, clean up some space for the tracks",
    'albumDoesntExsists': "Track's album does not exsist, failed to gather info"
}

def downloadImage(url, path, overwrite=OverwriteOption.DONT_OVERWRITE):
    if not path.is_file() or overwrite in [OverwriteOption.OVERWRITE, OverwriteOption.ONLY_TAGS, OverwriteOption.KEEP_BOTH]:
        try:
            image = get(url, headers={'User-Agent': USER_AGENT_HEADER}, timeout=30)
            image.raise_for_status()
            with open(path, 'wb') as f:
                f.write(image.content)
            return path
        except request_exception.HTTPError:
            if 'cdns-images.dzcdn.net' in url:
                urlBase = url[:url.rfind("/")+1]
                pictureUrl = url[len(urlBase):]
                pictureSize = int(pictureUrl[:pictureUrl.find("x")])
                if pictureSize > 1200:
                    logger.warn("Couldn't download "+str(pictureSize)+"x"+str(pictureSize)+" image, falling back to 1200x1200")
                    eventlet.sleep(1)
                    return  downloadImage(urlBase+pictureUrl.replace(str(pictureSize)+"x"+str(pictureSize), '1200x1200'), path, overwrite)
            logger.error("Image not found: "+url)
        except (request_exception.ConnectionError, request_exception.ChunkedEncodingError) as e:
            logger.error("Couldn't download Image, retrying in 5 seconds...: "+url+"\n")
            eventlet.sleep(5)
            return downloadImage(url, path, overwrite)
        except OSError as e:
            if e.errno == errno.ENOSPC:
                raise DownloadFailed("noSpaceLeft")
            else:
                logger.exception(f"Error while downloading an image, you should report this to the developers: {str(e)}")
        except Exception as e:
            logger.exception(f"Error while downloading an image, you should report this to the developers: {str(e)}")
        if path.is_file(): path.unlink()
        return None
    else:
        return path

def formatDate(date, template):
    elements = {
        'year': ['YYYY', 'YY', 'Y'],
        'month': ['MM', 'M'],
        'day': ['DD', 'D']
    }
    for element, placeholders in elements.items():
        for placeholder in placeholders:
            if placeholder in template:
                template = template.replace(placeholder, str(date[element]))
    return template

class DownloadJob:
    def __init__(self, dz, queueItem, interface=None):
        self.dz = dz
        self.interface = interface
        self.queueItem = queueItem
        self.settings = queueItem.settings
        self.bitrate = queueItem.bitrate
        self.downloadPercentage = 0
        self.lastPercentage = 0
        self.extrasPath = None
        self.playlistCoverName = None
        self.playlistURLs = []

    def start(self):
        if not self.queueItem.cancel:
            if isinstance(self.queueItem, QISingle):
                result = self.downloadWrapper(self.queueItem.single)
                if result: self.singleAfterDownload(result)
            elif isinstance(self.queueItem, QICollection):
                tracks = [None] * len(self.queueItem.collection)
                pool = eventlet.GreenPool(size=self.settings['queueConcurrency'])
                for pos, track in enumerate(self.queueItem.collection, start=0):
                    tracks[pos] = pool.spawn(self.downloadWrapper, track)
                pool.waitall()
                self.collectionAfterDownload(tracks)
        if self.interface:
            if self.queueItem.cancel:
                self.interface.send('currentItemCancelled', self.queueItem.uuid)
                self.interface.send("removedFromQueue", self.queueItem.uuid)
            else:
                self.interface.send("finishDownload", self.queueItem.uuid)
        return self.extrasPath

    def singleAfterDownload(self, result):
        if not self.extrasPath: self.extrasPath = Path(self.settings['downloadLocation'])

        # Save Album Cover
        if self.settings['saveArtwork'] and 'albumPath' in result:
            for image in result['albumURLs']:
                downloadImage(image['url'], result['albumPath'] / f"{result['albumFilename']}.{image['ext']}", self.settings['overwriteFile'])

        # Save Artist Artwork
        if self.settings['saveArtworkArtist'] and 'artistPath' in result:
            for image in result['artistURLs']:
                downloadImage(image['url'], result['artistPath'] / f"{result['artistFilename']}.{image['ext']}", self.settings['overwriteFile'])

        # Create searched logfile
        if self.settings['logSearched'] and 'searched' in result:
            with open(self.extrasPath / 'searched.txt', 'wb+') as f:
                orig = f.read().decode('utf-8')
                if not result['searched'] in orig:
                    if orig != "": orig += "\r\n"
                    orig += result['searched'] + "\r\n"
                f.write(orig.encode('utf-8'))
        # Execute command after download
        if self.settings['executeCommand'] != "":
            execute(self.settings['executeCommand'].replace("%folder%", quote(str(self.extrasPath))).replace("%filename%", quote(result['filename'])), shell=True)

    def collectionAfterDownload(self, tracks):
        if not self.extrasPath: self.extrasPath = Path(self.settings['downloadLocation'])
        playlist = [None] * len(tracks)
        errors = ""
        searched = ""

        for i in range(len(tracks)):
            result = tracks[i].wait()
            if not result: return None # Check if item is cancelled

            # Log errors to file
            if result.get('error'):
                if not result['error'].get('data'): result['error']['data'] = {'id': "0", 'title': 'Unknown', 'artist': 'Unknown'}
                errors += f"{result['error']['data']['id']} | {result['error']['data']['artist']} - {result['error']['data']['title']} | {result['error']['message']}\r\n"

            # Log searched to file
            if 'searched' in result: searched += result['searched'] + "\r\n"

            # Save Album Cover
            if self.settings['saveArtwork'] and 'albumPath' in result:
                for image in result['albumURLs']:
                    downloadImage(image['url'], result['albumPath'] / f"{result['albumFilename']}.{image['ext']}", self.settings['overwriteFile'])

            # Save Artist Artwork
            if self.settings['saveArtworkArtist'] and 'artistPath' in result:
                for image in result['artistURLs']:
                    downloadImage(image['url'], result['artistPath'] / f"{result['artistFilename']}.{image['ext']}", self.settings['overwriteFile'])

            # Save filename for playlist file
            playlist[i] = result.get('filename', "")

        # Create errors logfile
        if self.settings['logErrors'] and errors != "":
            with open(self.extrasPath / 'errors.txt', 'wb') as f:
                f.write(errors.encode('utf-8'))

        # Create searched logfile
        if self.settings['logSearched'] and searched != "":
            with open(self.extrasPath / 'searched.txt', 'wb') as f:
                f.write(searched.encode('utf-8'))

        # Save Playlist Artwork
        if self.settings['saveArtwork'] and self.playlistCoverName and not self.settings['tags']['savePlaylistAsCompilation']:
            for image in self.playlistURLs:
                downloadImage(image['url'], self.extrasPath / f"{self.playlistCoverName}.{image['ext']}", self.settings['overwriteFile'])

        # Create M3U8 File
        if self.settings['createM3U8File']:
            filename = settingsRegexPlaylistFile(self.settings['playlistFilenameTemplate'], self.queueItem, self.settings) or "playlist"
            with open(self.extrasPath / f'{filename}.m3u8', 'wb') as f:
                for line in playlist:
                    f.write((line + "\n").encode('utf-8'))

        # Execute command after download
        if self.settings['executeCommand'] != "":
            execute(self.settings['executeCommand'].replace("%folder%", quote(str(self.extrasPath))), shell=True)

    def download(self, trackAPI_gw, track=None):
        result = {}
        if self.queueItem.cancel: raise DownloadCancelled
        if trackAPI_gw['SNG_ID'] == "0": raise DownloadFailed("notOnDeezer")

        # Create Track object
        if not track:
            logger.info(f"[{trackAPI_gw['ART_NAME']} - {trackAPI_gw['SNG_TITLE']}] Getting the tags")
            try:
                track = Track(self.dz,
                              settings=self.settings,
                              trackAPI_gw=trackAPI_gw,
                              trackAPI=trackAPI_gw['_EXTRA_TRACK'] if '_EXTRA_TRACK' in trackAPI_gw else None,
                              albumAPI=trackAPI_gw['_EXTRA_ALBUM'] if '_EXTRA_ALBUM' in trackAPI_gw else None
                              )
            except AlbumDoesntExsists:
                raise DownloadError('albumDoesntExsists')
            if self.queueItem.cancel: raise DownloadCancelled

        # Check if track not yet encoded
        if track.MD5 == '':
            if track.fallbackId != "0":
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not yet encoded, using fallback id")
                newTrack = self.dz.get_track_gw(track.fallbackId)
                track.parseEssentialData(self.dz, newTrack)
                return self.download(trackAPI_gw, track)
            elif not track.searched and self.settings['fallbackSearch']:
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not yet encoded, searching for alternative")
                searchedId = self.dz.get_track_from_metadata(track.mainArtist['name'], track.title, track.album['title'])
                if searchedId != "0":
                    newTrack = self.dz.get_track_gw(searchedId)
                    track.parseEssentialData(self.dz, newTrack)
                    track.searched = True
                    if self.interface:
                        self.interface.send('queueUpdate', {
                            'uuid': self.queueItem.uuid,
                            'searchFallback': True,
                            'data': {
                                'id': track.id,
                                'title': track.title,
                                'artist': track.mainArtist['name']
                            },
                        })
                    return self.download(trackAPI_gw, track)
                else:
                    raise DownloadFailed("notEncodedNoAlternative")
            else:
                raise DownloadFailed("notEncoded")

        try:
            selectedFormat = self.getPreferredBitrate(track)
        except PreferredBitrateNotFound:
            if track.fallbackId != "0":
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not found at desired bitrate, using fallback id")
                newTrack = self.dz.get_track_gw(track.fallbackId)
                track.parseEssentialData(self.dz, newTrack)
                return self.download(trackAPI_gw, track)
            elif not track.searched and self.settings['fallbackSearch']:
                logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not found at desired bitrate, searching for alternative")
                searchedId = self.dz.get_track_from_metadata(track.mainArtist['name'], track.title, track.album['title'])
                if searchedId != "0":
                    newTrack = self.dz.get_track_gw(searchedId)
                    track.parseEssentialData(self.dz, newTrack)
                    track.searched = True
                    if self.interface:
                        self.interface.send('queueUpdate', {
                            'uuid': self.queueItem.uuid,
                            'searchFallback': True,
                            'data': {
                                'id': track.id,
                                'title': track.title,
                                'artist': track.mainArtist['name']
                            },
                        })
                    return self.download(trackAPI_gw, track)
                else:
                    raise DownloadFailed("wrongBitrateNoAlternative")
            else:
                raise DownloadFailed("wrongBitrate")
        except TrackNot360:
            raise DownloadFailed("no360RA")
        track.selectedFormat = selectedFormat

        imageQualityJPEG = f'000000-{self.settings["jpegImageQuality"]}-0-0.jpg'
        imageQualityPNG = 'none-100-0-0.png'
        imageQuality = imageQualityJPEG
        if self.settings['embeddedArtworkPNG']: imageQuality = imageQualityPNG

        if self.settings['tags']['savePlaylistAsCompilation'] and track.playlist:
            track.trackNumber = track.position
            track.discNumber = "1"
            track.album = {**track.album, **track.playlist}
            if 'picType' in track.playlist:
                track.playlist['picUrl'] = "https://e-cdns-images.dzcdn.net/images/{}/{}/{}x{}-{}".format(
                    track.playlist['picType'],
                    track.playlist['pic'],
                    self.settings['embeddedArtworkSize'], self.settings['embeddedArtworkSize'],
                    imageQuality
                )
            else:
                track.playlist['picUrl'] = track.playlist['pic']

            ext = track.playlist['picUrl'][-4:]
            if ext[0] != ".": ext = ".jpg"

            track.album['picPath'] = TEMPDIR / f"pl{trackAPI_gw['_EXTRA_PLAYLIST']['id']}_{self.settings['embeddedArtworkSize']}{ext}"
        else:
            if track.album['date']: track.date = track.album['date']
            track.album['picUrl'] = "https://e-cdns-images.dzcdn.net/images/cover/{}/{}x{}-{}".format(
                track.album['pic'],
                self.settings['embeddedArtworkSize'], self.settings['embeddedArtworkSize'],
                imageQuality
            )

            track.album['picPath'] = TEMPDIR / f"alb{track.album['id']}_{self.settings['embeddedArtworkSize']}{track.album['picUrl'][-4:]}"

        track.album['bitrate'] = selectedFormat

        track.dateString = formatDate(track.date, self.settings['dateFormat'])
        track.album['dateString'] = formatDate(track.album['date'], self.settings['dateFormat'])
        if track.playlist: track.playlist['dateString'] = formatDate(track.playlist['date'], self.settings['dateFormat'])

        # Check if user wants the feat in the title
        if str(self.settings['featuredToTitle']) == FeaturesOption.REMOVE_TITLE:
            track.title = track.getCleanTitle()
        elif str(self.settings['featuredToTitle']) == FeaturesOption.MOVE_TITLE:
            track.title = track.getFeatTitle()
        elif str(self.settings['featuredToTitle']) == FeaturesOption.REMOVE_TITLE_ALBUM:
            track.title = track.getCleanTitle()
            track.album['title'] = track.getCleanAlbumTitle()

        # Remove (Album Version) from tracks that have that
        if self.settings['removeAlbumVersion']:
            if "Album Version" in track.title:
                track.title = re.sub(r' ?\(Album Version\)', "", track.title).strip()

        # Change Title and Artists casing if needed
        if self.settings['titleCasing'] != "nothing":
            track.title = changeCase(track.title, self.settings['titleCasing'])
        if self.settings['artistCasing'] != "nothing":
            track.mainArtist['name'] = changeCase(track.mainArtist['name'], self.settings['artistCasing'])
            for i, artist in enumerate(track.artists):
                track.artists[i] = changeCase(artist, self.settings['artistCasing'])
            for type in track.artist:
                for i, artist in enumerate(track.artist[type]):
                    track.artist[type][i] = changeCase(artist, self.settings['artistCasing'])
            track.generateMainFeatStrings()

        # Generate artist tag
        if self.settings['tags']['multiArtistSeparator'] == "default":
            if str(self.settings['featuredToTitle']) == FeaturesOption.MOVE_TITLE:
                track.artistsString = ", ".join(track.artist['Main'])
            else:
                track.artistsString = ", ".join(track.artists)
        elif self.settings['tags']['multiArtistSeparator'] == "andFeat":
            track.artistsString = track.mainArtistsString
            if track.featArtistsString and str(self.settings['featuredToTitle']) != FeaturesOption.MOVE_TITLE:
                track.artistsString += " " + track.featArtistsString
        else:
            separator = self.settings['tags']['multiArtistSeparator']
            if str(self.settings['featuredToTitle']) == FeaturesOption.MOVE_TITLE:
                track.artistsString = separator.join(track.artist['Main'])
            else:
                track.artistsString = separator.join(track.artists)



        # Generate filename and filepath from metadata
        filename = generateFilename(track, self.settings, trackAPI_gw['FILENAME_TEMPLATE'])
        (filepath, artistPath, coverPath, extrasPath) = generateFilepath(track, self.settings)

        if self.queueItem.cancel: raise DownloadCancelled

        # Download and cache coverart
        logger.info(f"[{track.mainArtist['name']} - {track.title}] Getting the album cover")
        track.album['picPath'] = downloadImage(track.album['picUrl'], track.album['picPath'])

        # Save local album art
        if coverPath:
            result['albumURLs'] = []
            for format in self.settings['localArtworkFormat'].split(","):
                if format in ["png","jpg"]:
                    if self.settings['tags']['savePlaylistAsCompilation'] and track.playlist:
                        if track.playlist['picType']:
                            localImageQuality = imageQualityJPEG
                            if format == "png": localImageQuality = imageQualityPNG
                            url = "https://e-cdns-images.dzcdn.net/images/{}/{}/{}x{}-{}".format(
                                track.album['picType'],
                                track.album['pic'],
                                self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                                localImageQuality
                            )
                        else:
                            url = track.album['pic']
                            if format != "jpg": continue
                    else:
                        localImageQuality = imageQualityJPEG
                        if format == "png": localImageQuality = imageQualityPNG
                        url = "https://e-cdns-images.dzcdn.net/images/cover/{}/{}x{}-{}".format(
                            track.album['pic'],
                            self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                            localImageQuality
                        )
                    result['albumURLs'].append({'url': url, 'ext': format})
            result['albumPath'] = coverPath
            result['albumFilename'] = f"{settingsRegexAlbum(self.settings['coverImageTemplate'], track.album, self.settings, track.playlist)}"

        # Save artist art
        if artistPath:
            result['artistURLs'] = []
            for format in self.settings['localArtworkFormat'].split(","):
                if format in ["png","jpg"]:
                    url = ""
                    if track.album['mainArtist']['pic'] != "":
                        localImageQuality = imageQualityJPEG
                        if format == "png": localImageQuality = imageQualityPNG
                        url = "https://e-cdns-images.dzcdn.net/images/artist/{}/{}x{}-{}".format(
                            track.album['mainArtist']['pic'],
                            self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                            localImageQuality
                        )
                    elif format == "jpg": # Blank artist image is not available in PNG
                        url = "https://e-cdns-images.dzcdn.net/images/artist//{}x{}-{}".format(
                            self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                            imageQualityJPEG
                        )
                    if url: result['artistURLs'].append({'url': url, 'ext': format})
            result['artistPath'] = artistPath
            result['artistFilename'] = f"{settingsRegexArtist(self.settings['artistImageTemplate'], track.album['mainArtist'], self.settings)}"

        # Remove subfolders from filename and add it to filepath
        if pathSep in filename:
            tempPath = filename[:filename.rfind(pathSep)]
            filepath = filepath / tempPath
            filename = filename[filename.rfind(pathSep) + len(pathSep):]

        # Make sure the filepath exsists
        makedirs(filepath, exist_ok=True)
        writepath = filepath / f"{filename}{extensions[track.selectedFormat]}"

        # Save lyrics in lrc file
        if self.settings['syncedLyrics'] and track.lyrics['sync']:
            if not (filepath / f"{filename}.lrc").is_file() or self.settings['overwriteFile'] in [OverwriteOption.OVERWRITE, OverwriteOption.ONLY_TAGS]:
                with open(filepath / f"{filename}.lrc", 'wb') as f:
                    f.write(track.lyrics['sync'].encode('utf-8'))

        trackAlreadyDownloaded = writepath.is_file()

        # Don't overwrite and don't mind extension
        if not trackAlreadyDownloaded and self.settings['overwriteFile'] == OverwriteOption.DONT_CHECK_EXT:
            exts = ['.mp3', '.flac', '.opus', '.m4a']
            baseFilename = str(filepath / filename)
            for ext in exts:
                trackAlreadyDownloaded = Path(baseFilename+ext).is_file()
                if trackAlreadyDownloaded: break

        # Don't overwrite and keep both files
        if trackAlreadyDownloaded and self.settings['overwriteFile'] == OverwriteOption.KEEP_BOTH:
            baseFilename = str(filepath / filename)
            i = 1
            currentFilename = baseFilename+' ('+str(i)+')'+ extensions[track.selectedFormat]
            while Path(currentFilename).is_file():
                i += 1
                currentFilename = baseFilename+' ('+str(i)+')'+ extensions[track.selectedFormat]
            trackAlreadyDownloaded = False
            writepath = Path(currentFilename)

        if extrasPath:
            if not self.extrasPath: self.extrasPath = extrasPath
            result['filename'] = str(writepath)[len(str(extrasPath))+ len(pathSep):]

        # Save playlist cover
        if track.playlist:
            if not len(self.playlistURLs):
                if track.playlist['picType']:
                    for format in self.settings['localArtworkFormat'].split(","):
                        if format in ["png","jpg"]:
                            localImageQuality = imageQualityJPEG
                            if format == "png": localImageQuality = imageQualityPNG
                            url = "https://e-cdns-images.dzcdn.net/images/{}/{}/{}x{}-{}".format(
                                track.playlist['picType'],
                                track.playlist['pic'],
                                self.settings['localArtworkSize'], self.settings['localArtworkSize'],
                                localImageQuality
                            )
                            self.playlistURLs.append({'url': url, 'ext': format})
                else:
                    self.playlistURLs.append({'url': track.playlist['pic'], 'ext': 'jpg'})
            if not self.playlistCoverName:
                track.playlist['id'] = "pl_" + str(trackAPI_gw['_EXTRA_PLAYLIST']['id'])
                track.playlist['genre'] = ["Compilation", ]
                track.playlist['bitrate'] = selectedFormat
                track.playlist['dateString'] = formatDate(track.playlist['date'], self.settings['dateFormat'])
                self.playlistCoverName = f"{settingsRegexAlbum(self.settings['coverImageTemplate'], track.playlist, self.settings, track.playlist)}"

        if not trackAlreadyDownloaded or self.settings['overwriteFile'] == OverwriteOption.OVERWRITE:
            logger.info(f"[{track.mainArtist['name']} - {track.title}] Downloading the track")
            track.downloadUrl = self.dz.get_track_stream_url(track.id, track.MD5, track.mediaVersion, track.selectedFormat)

            def downloadMusic(track, trackAPI_gw):
                try:
                    with open(writepath, 'wb') as stream:
                        self.streamTrack(stream, track)
                except DownloadCancelled:
                    if writepath.is_file(): writepath.unlink()
                    raise DownloadCancelled
                except (request_exception.HTTPError, DownloadEmpty):
                    if writepath.is_file(): writepath.unlink()
                    if track.fallbackId != "0":
                        logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not available, using fallback id")
                        newTrack = self.dz.get_track_gw(track.fallbackId)
                        track.parseEssentialData(self.dz, newTrack)
                        return False
                    elif not track.searched and self.settings['fallbackSearch']:
                        logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not available, searching for alternative")
                        searchedId = self.dz.get_track_from_metadata(track.mainArtist['name'], track.title, track.album['title'])
                        if searchedId != "0":
                            newTrack = self.dz.get_track_gw(searchedId)
                            track.parseEssentialData(self.dz, newTrack)
                            track.searched = True
                            if self.interface:
                                self.interface.send('queueUpdate', {
                                    'uuid': self.queueItem.uuid,
                                    'searchFallback': True,
                                    'data': {
                                        'id': track.id,
                                        'title': track.title,
                                        'artist': track.mainArtist['name']
                                    },
                                })
                            return False
                        else:
                            raise DownloadFailed("notAvailableNoAlternative")
                    else:
                        raise DownloadFailed("notAvailable")
                except (request_exception.ConnectionError, request_exception.ChunkedEncodingError) as e:
                    if writepath.is_file(): writepath.unlink()
                    logger.warn(f"[{track.mainArtist['name']} - {track.title}] Error while downloading the track, trying again in 5s...")
                    eventlet.sleep(5)
                    return downloadMusic(track, trackAPI_gw)
                except OSError as e:
                    if e.errno == errno.ENOSPC:
                        raise DownloadFailed("noSpaceLeft")
                    else:
                        if writepath.is_file(): writepath.unlink()
                        logger.exception(f"[{track.mainArtist['name']} - {track.title}] Error while downloading the track, you should report this to the developers: {str(e)}")
                        raise e
                except Exception as e:
                    if writepath.is_file(): writepath.unlink()
                    logger.exception(f"[{track.mainArtist['name']} - {track.title}] Error while downloading the track, you should report this to the developers: {str(e)}")
                    raise e
                return True

            try:
                trackDownloaded = downloadMusic(track, trackAPI_gw)
            except Exception as e:
                raise e

            if not trackDownloaded: return self.download(trackAPI_gw, track)
        else:
            logger.info(f"[{track.mainArtist['name']} - {track.title}] Skipping track as it's already downloaded")
            self.completeTrackPercentage()

        # Adding tags
        if (not trackAlreadyDownloaded or self.settings['overwriteFile'] in [OverwriteOption.ONLY_TAGS, OverwriteOption.OVERWRITE]) and not track.localTrack:
            logger.info(f"[{track.mainArtist['name']} - {track.title}] Applying tags to the track")
            if track.selectedFormat in [TrackFormats.MP3_320, TrackFormats.MP3_128, TrackFormats.DEFAULT]:
                tagID3(writepath, track, self.settings['tags'])
            elif track.selectedFormat ==  TrackFormats.FLAC:
                try:
                    tagFLAC(writepath, track, self.settings['tags'])
                except (FLACNoHeaderError, FLACError):
                    if writepath.is_file(): writepath.unlink()
                    logger.warn(f"[{track.mainArtist['name']} - {track.title}] Track not available in FLAC, falling back if necessary")
                    self.removeTrackPercentage()
                    track.filesizes['FILESIZE_FLAC'] = "0"
                    track.filesizes['FILESIZE_FLAC_TESTED'] = True
                    return self.download(trackAPI_gw, track)

        if track.searched: result['searched'] = f"{track.mainArtist['name']} - {track.title}"
        logger.info(f"[{track.mainArtist['name']} - {track.title}] Track download completed\n{str(writepath)}")
        self.queueItem.downloaded += 1
        self.queueItem.files.append(str(writepath))
        self.queueItem.extrasPath = str(self.extrasPath)
        if self.interface:
            self.interface.send("updateQueue", {'uuid': self.queueItem.uuid, 'downloaded': True, 'downloadPath': str(writepath), 'extrasPath': str(self.extrasPath)})
        return result

    def getPreferredBitrate(self, track):
        if track.localTrack: return TrackFormats.LOCAL

        shouldFallback = self.settings['fallbackBitrate']
        falledBack = False

        formats_non_360 = {
            TrackFormats.FLAC: "FLAC",
            TrackFormats.MP3_320: "MP3_320",
            TrackFormats.MP3_128: "MP3_128",
        }
        formats_360 = {
            TrackFormats.MP4_RA3: "MP4_RA3",
            TrackFormats.MP4_RA2: "MP4_RA2",
            TrackFormats.MP4_RA1: "MP4_RA1",
        }

        is360format = int(self.bitrate) in formats_360

        if not shouldFallback:
            formats = formats_360
            formats.update(formats_non_360)
        elif is360format:
            formats = formats_360
        else:
            formats = formats_non_360

        for formatNumber, formatName in formats.items():
            if formatNumber <= int(self.bitrate):
                if f"FILESIZE_{formatName}" in track.filesizes:
                    if int(track.filesizes[f"FILESIZE_{formatName}"]) != 0: return formatNumber
                    if not track.filesizes[f"FILESIZE_{formatName}_TESTED"]:
                        request = requests.head(
                            self.dz.get_track_stream_url(track.id, track.MD5, track.mediaVersion, formatNumber),
                            headers={'User-Agent': USER_AGENT_HEADER},
                            timeout=30
                        )
                        try:
                            request.raise_for_status()
                            return formatNumber
                        except request_exception.HTTPError: # if the format is not available, Deezer returns a 403 error
                            pass
                if not shouldFallback:
                    raise PreferredBitrateNotFound
                else:
                    if not falledBack:
                        falledBack = True
                        logger.info(f"[{track.mainArtist['name']} - {track.title}] Fallback to lower bitrate")
                        if self.interface:
                            self.interface.send('queueUpdate', {
                                'uuid': self.queueItem.uuid,
                                'bitrateFallback': True,
                                'data': {
                                    'id': track.id,
                                    'title': track.title,
                                    'artist': track.mainArtist['name']
                                },
                            })
        if is360format: raise TrackNot360
        return TrackFormats.DEFAULT

    def streamTrack(self, stream, track, start=0):
        if self.queueItem.cancel: raise DownloadCancelled

        headers=dict(self.dz.http_headers)
        if range != 0: headers['Range'] = f'bytes={start}-'
        chunkLength = start
        percentage = 0

        itemName = f"[{track.mainArtist['name']} - {track.title}]"

        try:
            with self.dz.session.get(track.downloadUrl, headers=headers, stream=True, timeout=10) as request:
                request.raise_for_status()

                blowfish_key = str.encode(self.dz._get_blowfish_key(str(track.id)))

                complete = int(request.headers["Content-Length"])
                if complete == 0: raise DownloadEmpty
                if start != 0:
                    responseRange = request.headers["Content-Range"]
                    logger.info(f'{itemName} downloading range {responseRange}')
                else:
                    logger.info(f'{itemName} downloading {complete} bytes')

                for chunk in request.iter_content(2048 * 3):
                    if self.queueItem.cancel: raise DownloadCancelled

                    if len(chunk) >= 2048:
                        chunk = Blowfish.new(blowfish_key, Blowfish.MODE_CBC, b"\x00\x01\x02\x03\x04\x05\x06\x07").decrypt(chunk[0:2048]) + chunk[2048:]

                    stream.write(chunk)
                    chunkLength += len(chunk)

                    if isinstance(self.queueItem, QISingle):
                        percentage = (chunkLength / (complete + start)) * 100
                        self.downloadPercentage = percentage
                    else:
                        chunkProgres = (len(chunk) / (complete + start)) / self.queueItem.size * 100
                        self.downloadPercentage += chunkProgres
                    self.updatePercentage()

        except SSLError as e:
            logger.info(f'{itemName} retrying from byte {chunkLength}')
            return self.streamTrack(stream, track, chunkLength)
        except (request_exception.ConnectionError, requests.exceptions.ReadTimeout):
            eventlet.sleep(2)
            return self.streamTrack(stream, track, start)

    def updatePercentage(self):
        if round(self.downloadPercentage) != self.lastPercentage and round(self.downloadPercentage) % 2 == 0:
            self.lastPercentage = round(self.downloadPercentage)
            self.queueItem.progress = self.lastPercentage
            if self.interface: self.interface.send("updateQueue", {'uuid': self.queueItem.uuid, 'progress': self.lastPercentage})

    def completeTrackPercentage(self):
        if isinstance(self.queueItem, QISingle):
            self.downloadPercentage = 100
        else:
            self.downloadPercentage += (1 / self.queueItem.size) * 100
        self.updatePercentage()

    def removeTrackPercentage(self):
        if isinstance(self.queueItem, QISingle):
            self.downloadPercentage = 0
        else:
            self.downloadPercentage -= (1 / self.queueItem.size) * 100
        self.updatePercentage()

    def downloadWrapper(self, trackAPI_gw):
        track = {
            'id': trackAPI_gw['SNG_ID'],
            'title': trackAPI_gw['SNG_TITLE'].strip(),
            'artist': trackAPI_gw['ART_NAME']
        }
        if trackAPI_gw.get('VERSION') and trackAPI_gw['VERSION'] not in trackAPI_gw['SNG_TITLE']:
            track['title'] += f" {trackAPI_gw['VERSION']}".strip()

        try:
            result = self.download(trackAPI_gw)
        except DownloadCancelled:
            return None
        except DownloadFailed as error:
            logger.error(f"[{track['artist']} - {track['title']}] {error.message}")
            result = {'error': {
                        'message': error.message,
                        'errid': error.errid,
                        'data': track
                    }}
        except Exception as e:
            logger.exception(f"[{track['artist']} - {track['title']}] {str(e)}")
            result = {'error': {
                        'message': str(e),
                        'data': track
                    }}

        if 'error' in result:
            self.completeTrackPercentage()
            self.queueItem.failed += 1
            self.queueItem.errors.append(result['error'])
            if self.interface:
                error = result['error']
                self.interface.send("updateQueue", {
                    'uuid': self.queueItem.uuid,
                    'failed': True,
                    'data': error['data'],
                    'error': error['message'],
                    'errid': error['errid'] if 'errid' in error else None
                })
        return result

class DownloadError(Exception):
    """Base class for exceptions in this module."""
    pass

class DownloadFailed(DownloadError):
    def __init__(self, errid):
        self.errid = errid
        self.message = errorMessages[self.errid]

class DownloadCancelled(DownloadError):
    pass

class DownloadEmpty(DownloadError):
    pass

class PreferredBitrateNotFound(DownloadError):
    pass

class TrackNot360(DownloadError):
    pass



from deemix.app.downloadjob import DownloadJob
from deemix.utils import getIDFromLink, getTypeFromLink, getBitrateInt
from deemix.api.deezer import APIError, LyricsStatus
from spotipy.exceptions import SpotifyException
from deemix.app.queueitem import QueueItem, QISingle, QICollection, QIConvertable
import logging
from pathlib import Path
import json
from os import remove
import eventlet
urlopen = eventlet.import_patched('urllib.request').urlopen

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

class QueueManager:
    def __init__(self, spotifyHelper=None):
        self.queue = []
        self.queueList = {}
        self.queueComplete = []
        self.currentItem = ""
        self.sp = spotifyHelper

    def generateTrackQueueItem(self, dz, id, settings, bitrate, trackAPI=None, albumAPI=None):
        # Check if is an isrc: url
        if str(id).startswith("isrc"):
            try:
                trackAPI = dz.get_track(id)
            except APIError as e:
                e = json.loads(str(e))
                return QueueError("https://deezer.com/track/"+str(id), f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")
            if 'id' in trackAPI and 'title' in trackAPI:
                id = trackAPI['id']
            else:
                return QueueError("https://deezer.com/track/"+str(id), "Track ISRC is not available on deezer", "ISRCnotOnDeezer")

        # Get essential track info
        try:
            trackAPI_gw = dz.get_track_gw(id)
        except APIError as e:
            e = json.loads(str(e))
            message = "Wrong URL"
            if "DATA_ERROR" in e: message += f": {e['DATA_ERROR']}"
            return QueueError("https://deezer.com/track/"+str(id), message)

        if albumAPI: trackAPI_gw['_EXTRA_ALBUM'] = albumAPI
        if trackAPI: trackAPI_gw['_EXTRA_TRACK'] = trackAPI

        if settings['createSingleFolder']:
            trackAPI_gw['FILENAME_TEMPLATE'] = settings['albumTracknameTemplate']
        else:
            trackAPI_gw['FILENAME_TEMPLATE'] = settings['tracknameTemplate']

        trackAPI_gw['SINGLE_TRACK'] = True

        title = trackAPI_gw['SNG_TITLE'].strip()
        if trackAPI_gw.get('VERSION') and trackAPI_gw['VERSION'] not in trackAPI_gw['SNG_TITLE']:
            title += f" {trackAPI_gw['VERSION']}".strip()
        explicit = bool(int(trackAPI_gw.get('EXPLICIT_LYRICS', 0)))

        return QISingle(
            id=id,
            bitrate=bitrate,
            title=title,
            artist=trackAPI_gw['ART_NAME'],
            cover=f"https://e-cdns-images.dzcdn.net/images/cover/{trackAPI_gw['ALB_PICTURE']}/75x75-000000-80-0-0.jpg",
            explicit=explicit,
            type='track',
            settings=settings,
            single=trackAPI_gw,
        )

    def generateAlbumQueueItem(self, dz, id, settings, bitrate):
        # Get essential album info
        try:
            albumAPI = dz.get_album(id)
        except APIError as e:
            e = json.loads(str(e))
            return QueueError("https://deezer.com/album/"+str(id), f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

        if str(id).startswith('upc'): id = albumAPI['id']

        # Get extra info about album
        # This saves extra api calls when downloading
        albumAPI_gw = dz.get_album_gw(id)
        albumAPI['nb_disk'] = albumAPI_gw['NUMBER_DISK']
        albumAPI['copyright'] = albumAPI_gw['COPYRIGHT']

        # If the album is a single download as a track
        if albumAPI['nb_tracks'] == 1:
            return self.generateTrackQueueItem(dz, albumAPI['tracks']['data'][0]['id'], settings, bitrate, albumAPI=albumAPI)

        tracksArray = dz.get_album_tracks_gw(id)

        if albumAPI['cover_small'] != None:
            cover = albumAPI['cover_small'][:-24] + '/75x75-000000-80-0-0.jpg'
        else:
            cover = f"https://e-cdns-images.dzcdn.net/images/cover/{albumAPI_gw['ALB_PICTURE']}/75x75-000000-80-0-0.jpg"

        totalSize = len(tracksArray)
        albumAPI['nb_tracks'] = totalSize
        collection = []
        for pos, trackAPI in enumerate(tracksArray, start=1):
            trackAPI['_EXTRA_ALBUM'] = albumAPI
            trackAPI['POSITION'] = pos
            trackAPI['SIZE'] = totalSize
            trackAPI['FILENAME_TEMPLATE'] = settings['albumTracknameTemplate']
            collection.append(trackAPI)

        explicit = albumAPI_gw['EXPLICIT_ALBUM_CONTENT']['EXPLICIT_LYRICS_STATUS'] in [LyricsStatus.EXPLICIT, LyricsStatus.PARTIALLY_EXPLICIT]

        return QICollection(
            id=id,
            bitrate=bitrate,
            title=albumAPI['title'],
            artist=albumAPI['artist']['name'],
            cover=cover,
            explicit=explicit,
            size=totalSize,
            type='album',
            settings=settings,
            collection=collection,
        )

    def generatePlaylistQueueItem(self, dz, id, settings, bitrate):
        # Get essential playlist info
        try:
            playlistAPI = dz.get_playlist(id)
        except:
            playlistAPI = None
        # Fallback to gw api if the playlist is private
        if not playlistAPI:
            try:
                playlistAPI = dz.get_playlist_gw(id)
            except APIError as e:
                e = json.loads(str(e))
                message = "Wrong URL"
                if "DATA_ERROR" in e:
                    message += f": {e['DATA_ERROR']}"
                return QueueError("https://deezer.com/playlist/"+str(id), message)

        # Check if private playlist and owner
        if not playlistAPI['public'] and playlistAPI['creator']['id'] != str(dz.user['id']):
            logger.warn("You can't download others private playlists.")
            return QueueError("https://deezer.com/playlist/"+str(id), "You can't download others private playlists.", "notYourPrivatePlaylist")

        playlistTracksAPI = dz.get_playlist_tracks_gw(id)
        playlistAPI['various_artist'] = dz.get_artist(5080) # Useful for save as compilation

        totalSize = len(playlistTracksAPI)
        playlistAPI['nb_tracks'] = totalSize
        collection = []
        for pos, trackAPI in enumerate(playlistTracksAPI, start=1):
            if trackAPI['EXPLICIT_TRACK_CONTENT'].get('EXPLICIT_LYRICS_STATUS', LyricsStatus.UNKNOWN) in [LyricsStatus.EXPLICIT, LyricsStatus.PARTIALLY_EXPLICIT]:
                playlistAPI['explicit'] = True
            trackAPI['_EXTRA_PLAYLIST'] = playlistAPI
            trackAPI['POSITION'] = pos
            trackAPI['SIZE'] = totalSize
            trackAPI['FILENAME_TEMPLATE'] = settings['playlistTracknameTemplate']
            collection.append(trackAPI)
        if not 'explicit' in playlistAPI:
            playlistAPI['explicit'] = False

        return QICollection(
            id=id,
            bitrate=bitrate,
            title=playlistAPI['title'],
            artist=playlistAPI['creator']['name'],
            cover=playlistAPI['picture_small'][:-24] + '/75x75-000000-80-0-0.jpg',
            explicit=playlistAPI['explicit'],
            size=totalSize,
            type='playlist',
            settings=settings,
            collection=collection,
        )

    def generateArtistQueueItem(self, dz, id, settings, bitrate, interface=None):
        # Get essential artist info
        try:
            artistAPI = dz.get_artist(id)
        except APIError as e:
            e = json.loads(str(e))
            return QueueError("https://deezer.com/artist/"+str(id), f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

        if interface: interface.send("startAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})

        artistAPITracks = dz.get_artist_albums(id)
        albumList = []
        for album in artistAPITracks['data']:
            albumList.append(self.generateAlbumQueueItem(dz, album['id'], settings, bitrate))

        if interface: interface.send("finishAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})
        return albumList

    def generateArtistDiscographyQueueItem(self, dz, id, settings, bitrate, interface=None):
        # Get essential artist info
        try:
            artistAPI = dz.get_artist(id)
        except APIError as e:
            e = json.loads(str(e))
            return QueueError("https://deezer.com/artist/"+str(id)+"/discography", f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

        if interface: interface.send("startAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})

        artistDiscographyAPI = dz.get_artist_discography_gw(id, 100)
        artistDiscographyAPI.pop('all', None) # all contains albums and singles, so its all duplicates. This removes them
        albumList = []
        for type in artistDiscographyAPI:
            for album in artistDiscographyAPI[type]:
                albumList.append(self.generateAlbumQueueItem(dz, album['id'], settings, bitrate))

        if interface: interface.send("finishAddingArtist", {'name': artistAPI['name'], 'id': artistAPI['id']})
        return albumList

    def generateArtistTopQueueItem(self, dz, id, settings, bitrate, interface=None):
        # Get essential artist info
        try:
            artistAPI = dz.get_artist(id)
        except APIError as e:
            e = json.loads(str(e))
            return QueueError("https://deezer.com/artist/"+str(id)+"/top_track", f"Wrong URL: {e['type']+': ' if 'type' in e else ''}{e['message'] if 'message' in e else ''}")

        # Emulate the creation of a playlist
        # Can't use generatePlaylistQueueItem as this is not a real playlist
        playlistAPI = {
            'id': str(artistAPI['id'])+"_top_track",
            'title': artistAPI['name']+" - Top Tracks",
            'description': "Top Tracks for "+artistAPI['name'],
            'duration': 0,
            'public': True,
            'is_loved_track': False,
            'collaborative': False,
            'nb_tracks': 0,
            'fans': artistAPI['nb_fan'],
            'link': "https://www.deezer.com/artist/"+str(artistAPI['id'])+"/top_track",
            'share': None,
            'picture': artistAPI['picture'],
            'picture_small': artistAPI['picture_small'],
            'picture_medium': artistAPI['picture_medium'],
            'picture_big': artistAPI['picture_big'],
            'picture_xl': artistAPI['picture_xl'],
            'checksum': None,
            'tracklist': "https://api.deezer.com/artist/"+str(artistAPI['id'])+"/top",
            'creation_date': "XXXX-00-00",
            'creator': {
                'id': "art_"+str(artistAPI['id']),
                'name': artistAPI['name'],
                'type': "user"
            },
            'type': "playlist"
        }

        artistTopTracksAPI_gw = dz.get_artist_toptracks_gw(id)
        playlistAPI['various_artist'] = dz.get_artist(5080) # Useful for save as compilation

        totalSize = len(artistTopTracksAPI_gw)
        playlistAPI['nb_tracks'] = totalSize
        collection = []
        for pos, trackAPI in enumerate(artistTopTracksAPI_gw, start=1):
            if trackAPI['EXPLICIT_TRACK_CONTENT'].get('EXPLICIT_LYRICS_STATUS', LyricsStatus.UNKNOWN) in [LyricsStatus.EXPLICIT, LyricsStatus.PARTIALLY_EXPLICIT]:
                playlistAPI['explicit'] = True
            trackAPI['_EXTRA_PLAYLIST'] = playlistAPI
            trackAPI['POSITION'] = pos
            trackAPI['SIZE'] = totalSize
            trackAPI['FILENAME_TEMPLATE'] = settings['playlistTracknameTemplate']
            collection.append(trackAPI)
        if not 'explicit' in playlistAPI:
            playlistAPI['explicit'] = False

        return QICollection(
            id=id,
            bitrate=bitrate,
            title=playlistAPI['title'],
            artist=playlistAPI['creator']['name'],
            cover=playlistAPI['picture_small'][:-24] + '/75x75-000000-80-0-0.jpg',
            explicit=playlistAPI['explicit'],
            size=totalSize,
            type='playlist',
            settings=settings,
            collection=collection,
        )

    def generateQueueItem(self, dz, url, settings, bitrate=None, interface=None):
        bitrate = getBitrateInt(bitrate) or settings['maxBitrate']
        if 'deezer.page.link' in url: url = urlopen(url).url

        type = getTypeFromLink(url)
        id = getIDFromLink(url, type)
        if type == None or id == None:
            logger.warn("URL not recognized")
            return QueueError(url, "URL not recognized", "invalidURL")

        if type == "track":
            return self.generateTrackQueueItem(dz, id, settings, bitrate)
        elif type == "album":
            return self.generateAlbumQueueItem(dz, id, settings, bitrate)
        elif type == "playlist":
            return self.generatePlaylistQueueItem(dz, id, settings, bitrate)
        elif type == "artist":
            return self.generateArtistQueueItem(dz, id, settings, bitrate, interface=interface)
        elif type == "artistdiscography":
            return self.generateArtistDiscographyQueueItem(dz, id, settings, bitrate, interface=interface)
        elif type == "artisttop":
            return self.generateArtistTopQueueItem(dz, id, settings, bitrate, interface=interface)
        elif type.startswith("spotify") and self.sp:
            if not self.sp.spotifyEnabled:
                logger.warn("Spotify Features is not setted up correctly.")
                return QueueError(url, "Spotify Features is not setted up correctly.", "spotifyDisabled")

            if type == "spotifytrack":
                try:
                    (track_id, trackAPI, _) = self.sp.get_trackid_spotify(dz, id, settings['fallbackSearch'])
                except SpotifyException as e:
                    return QueueError(url, "Wrong URL: "+e.msg[e.msg.find('\n')+2:])
                except Exception as e:
                    return QueueError(url, "Something went wrong: "+str(e))

                if track_id != "0":
                    return self.generateTrackQueueItem(dz, track_id, settings, bitrate, trackAPI=trackAPI)
                else:
                    logger.warn("Track not found on deezer!")
                    return QueueError(url, "Track not found on deezer!", "trackNotOnDeezer")

            elif type == "spotifyalbum":
                try:
                    album_id = self.sp.get_albumid_spotify(dz, id)
                except SpotifyException as e:
                    return QueueError(url, "Wrong URL: "+e.msg[e.msg.find('\n')+2:])
                except Exception as e:
                    return QueueError(url, "Something went wrong: "+str(e))

                if album_id != "0":
                    return self.generateAlbumQueueItem(dz, album_id, settings, bitrate)
                else:
                    logger.warn("Album not found on deezer!")
                    return QueueError(url, "Album not found on deezer!", "albumNotOnDeezer")

            elif type == "spotifyplaylist":
                try:
                    return self.sp.generate_playlist_queueitem(dz, id, bitrate, settings)
                except SpotifyException as e:
                    return QueueError(url, "Wrong URL: "+e.msg[e.msg.find('\n')+2:])
                except Exception as e:
                    return QueueError(url, "Something went wrong: "+str(e))
        else:
            logger.warn("URL not supported yet")
            return QueueError(url, "URL not supported yet", "unsupportedURL")

    def addToQueue(self, dz, url, settings, bitrate=None, interface=None, ack=None):
        if not dz.logged_in:
            if interface: interface.send("loginNeededToDownload")
            return False

        def parseLink(link):
            link = link.strip()
            if link == "": return False
            logger.info("Generating queue item for: "+link)
            item = self.generateQueueItem(dz, link, settings, bitrate, interface=interface)

            # Add ack to all items
            if type(item) is list:
                for i in item:
                    if isinstance(i, QueueItem):
                        i.ack = ack
            elif isinstance(item, QueueItem):
                item.ack = ack
            return item

        if type(url) is list:
            queueItem = []
            for link in url:
                item = parseLink(link)
                if not item: continue
                if type(item) is list:
                    queueItem += item
                else:
                    queueItem.append(item)
            if not len(queueItem):
                return False
        else:
            queueItem = parseLink(url)
            if not queueItem:
                return False

        def processQueueItem(item, silent=False):
            if isinstance(item, QueueError):
                logger.error(f"[{item.link}] {item.message}")
                if interface and not silent: interface.send("queueError", item.toDict())
                return False
            if item.uuid in list(self.queueList.keys()):
                logger.warn(f"[{item.uuid}] Already in queue, will not be added again.")
                if interface and not silent: interface.send("alreadyInQueue", {'uuid': item.uuid, 'title': item.title})
                return False
            self.queue.append(item.uuid)
            self.queueList[item.uuid] = item
            logger.info(f"[{item.uuid}] Added to queue.")
            return True

        if type(queueItem) is list:
            slimmedItems = []
            for item in queueItem:
                if processQueueItem(item, silent=True):
                    slimmedItems.append(item.getSlimmedItem())
                else:
                    continue
            if not len(slimmedItems):
                return False
            if interface: interface.send("addedToQueue", slimmedItems)
        else:
            if processQueueItem(queueItem):
                if interface: interface.send("addedToQueue", queueItem.getSlimmedItem())
            else:
                return False
        self.nextItem(dz, interface)
        return True

    def nextItem(self, dz, interface=None):
        # Check that nothing is already downloading and
        # that the queue is not empty
        if self.currentItem != "": return None
        if not len(self.queue): return None

        self.currentItem = self.queue.pop(0)

        if isinstance(self.queueList[self.currentItem], QIConvertable) and self.queueList[self.currentItem].extra:
            logger.info(f"[{self.currentItem}] Converting tracks to deezer.")
            self.sp.convert_spotify_playlist(dz, self.queueList[self.currentItem], interface=interface)
            logger.info(f"[{self.currentItem}] Tracks converted.")

        if interface: interface.send("startDownload", self.currentItem)
        logger.info(f"[{self.currentItem}] Started downloading.")

        DownloadJob(dz, self.queueList[self.currentItem], interface).start()

        if self.queueList[self.currentItem].cancel:
            del self.queueList[self.currentItem]
        else:
            self.queueComplete.append(self.currentItem)
        logger.info(f"[{self.currentItem}] Finished downloading.")
        self.currentItem = ""
        self.nextItem(dz, interface)

    def getQueue(self):
        return (self.queue, self.queueComplete, self.slimQueueList(), self.currentItem)

    def saveQueue(self, configFolder):
        if len(self.queueList) > 0:
            if self.currentItem != "":
                self.queue.insert(0, self.currentItem)
            with open(Path(configFolder) / 'queue.json', 'w') as f:
                json.dump({
                    'queue': self.queue,
                    'queueComplete': self.queueComplete,
                    'queueList': self.exportQueueList()
                }, f)

    def exportQueueList(self):
        queueList = {}
        for uuid in self.queueList:
            if uuid in self.queue:
                queueList[uuid] = self.queueList[uuid].getResettedItem()
            else:
                queueList[uuid] = self.queueList[uuid].toDict()
        return queueList

    def slimQueueList(self):
        queueList = {}
        for uuid in self.queueList:
            queueList[uuid] = self.queueList[uuid].getSlimmedItem()
        return queueList

    def loadQueue(self, configFolder, settings, interface=None):
        configFolder = Path(configFolder)
        if (configFolder / 'queue.json').is_file() and not len(self.queue):
            if interface: interface.send('restoringQueue')
            with open(configFolder / 'queue.json', 'r') as f:
                qd = json.load(f)
            remove(configFolder / 'queue.json')
            self.restoreQueue(qd['queue'], qd['queueComplete'], qd['queueList'], settings)
            if interface:
                interface.send('init_downloadQueue', {
                    'queue': self.queue,
                    'queueComplete': self.queueComplete,
                    'queueList': self.slimQueueList(),
                    'restored': True
                })

    def restoreQueue(self, queue, queueComplete, queueList, settings):
        self.queue = queue
        self.queueComplete = queueComplete
        self.queueList = {}
        for uuid in queueList:
            if 'single' in queueList[uuid]:
                self.queueList[uuid] = QISingle(queueItemDict = queueList[uuid])
            if 'collection' in queueList[uuid]:
                self.queueList[uuid] = QICollection(queueItemDict = queueList[uuid])
            if '_EXTRA' in queueList[uuid]:
                self.queueList[uuid] = QIConvertable(queueItemDict = queueList[uuid])
            self.queueList[uuid].settings = settings

    def removeFromQueue(self, uuid, interface=None):
        if uuid == self.currentItem:
            if interface: interface.send("cancellingCurrentItem", uuid)
            self.queueList[uuid].cancel = True
            return
        if uuid in self.queue:
            self.queue.remove(uuid)
        elif uuid in self.queueComplete:
            self.queueComplete.remove(uuid)
        else:
            return
        del self.queueList[uuid]
        if interface: interface.send("removedFromQueue", uuid)


    def cancelAllDownloads(self, interface=None):
        self.queue = []
        self.queueComplete = []
        if self.currentItem != "":
            if interface: interface.send("cancellingCurrentItem", self.currentItem)
            self.queueList[self.currentItem].cancel = True
        for uuid in list(self.queueList.keys()):
            if uuid != self.currentItem: del self.queueList[uuid]
        if interface: interface.send("removedAllDownloads", self.currentItem)


    def removeFinishedDownloads(self, interface=None):
        for uuid in self.queueComplete:
            del self.queueList[uuid]
        self.queueComplete = []
        if interface: interface.send("removedFinishedDownloads")

class QueueError:
    def __init__(self, link, message, errid=None):
        self.link = link
        self.message = message
        self.errid = errid

    def toDict(self):
        return {
            'link': self.link,
            'error': self.message,
            'errid': self.errid
        }



import eventlet
import json
from pathlib import Path

eventlet.import_patched('requests.adapters')

spotipy = eventlet.import_patched('spotipy')
SpotifyClientCredentials = spotipy.oauth2.SpotifyClientCredentials
from deemix.utils.localpaths import getConfigFolder
from deemix.app.queueitem import QIConvertable

emptyPlaylist = {
    'collaborative': False,
    'description': "",
    'external_urls': {'spotify': None},
    'followers': {'total': 0, 'href': None},
    'id': None,
    'images': [],
    'name': "Something went wrong",
    'owner': {
        'display_name': "Error",
        'id': None
    },
    'public': True,
    'tracks' : [],
    'type': 'playlist',
    'uri': None
}

class SpotifyHelper:
    def __init__(self, configFolder=None):
        self.credentials = {}
        self.spotifyEnabled = False
        self.sp = None
        self.configFolder = configFolder

        # Make sure config folder exsists
        if not self.configFolder:
            self.configFolder = getConfigFolder()
        self.configFolder = Path(self.configFolder)
        if not self.configFolder.is_dir():
            self.configFolder.mkdir()

        # Make sure authCredentials exsits
        if not (self.configFolder / 'authCredentials.json').is_file():
            with open(self.configFolder / 'authCredentials.json', 'w') as f:
                json.dump({'clientId': "", 'clientSecret': ""}, f, indent=2)

        # Load spotify id and secret and check if they are usable
        with open(self.configFolder / 'authCredentials.json', 'r') as credentialsFile:
            self.credentials = json.load(credentialsFile)
        self.checkCredentials()
        self.checkValidCache()

    def checkValidCache(self):
        if (self.configFolder / 'spotifyCache.json').is_file():
            with open(self.configFolder / 'spotifyCache.json', 'r') as spotifyCache:
                try:
                    cache = json.load(spotifyCache)
                except Exception as e:
                    print(str(e))
                    (self.configFolder / 'spotifyCache.json').unlink()
                    return
            # Remove old versions of cache
            if len(cache['tracks'].values()) and isinstance(list(cache['tracks'].values())[0], int) or \
               len(cache['albums'].values()) and isinstance(list(cache['albums'].values())[0], int):
                (self.configFolder / 'spotifyCache.json').unlink()

    def checkCredentials(self):
        if self.credentials['clientId'] == "" or self.credentials['clientSecret'] == "":
            spotifyEnabled = False
        else:
            try:
                client_credentials_manager = SpotifyClientCredentials(client_id=self.credentials['clientId'],
                                                                      client_secret=self.credentials['clientSecret'])
                self.sp = spotipy.Spotify(client_credentials_manager=client_credentials_manager)
                self.sp.user_playlists('spotify')
                self.spotifyEnabled = True
            except Exception as e:
                self.spotifyEnabled = False
        return self.spotifyEnabled

    def getCredentials(self):
        return self.credentials

    def setCredentials(self, spotifyCredentials):
        # Remove extra spaces, just to be sure
        spotifyCredentials['clientId'] = spotifyCredentials['clientId'].strip()
        spotifyCredentials['clientSecret'] = spotifyCredentials['clientSecret'].strip()

        # Save them to disk
        with open(self.configFolder / 'authCredentials.json', 'w') as f:
            json.dump(spotifyCredentials, f, indent=2)

        # Check if they are usable
        self.credentials = spotifyCredentials
        self.checkCredentials()

    # Converts spotify API playlist structure to deezer's playlist structure
    def _convert_playlist_structure(self, spotify_obj):
        if len(spotify_obj['images']):
            url = spotify_obj['images'][0]['url']
        else:
            url = False
        deezer_obj = {
            'checksum': spotify_obj['snapshot_id'],
            'collaborative': spotify_obj['collaborative'],
            'creation_date': "XXXX-00-00",
            'creator': {
                'id': spotify_obj['owner']['id'],
                'name': spotify_obj['owner']['display_name'],
                'tracklist': spotify_obj['owner']['href'],
                'type': "user"
            },
            'description': spotify_obj['description'],
            'duration': 0,
            'fans': spotify_obj['followers']['total'] if 'followers' in spotify_obj else 0,
            'id': spotify_obj['id'],
            'is_loved_track': False,
            'link': spotify_obj['external_urls']['spotify'],
            'nb_tracks': spotify_obj['tracks']['total'],
            'picture': url,
            'picture_small': url,
            'picture_medium': url,
            'picture_big': url,
            'picture_xl': url,
            'public': spotify_obj['public'],
            'share': spotify_obj['external_urls']['spotify'],
            'title': spotify_obj['name'],
            'tracklist': spotify_obj['tracks']['href'],
            'type': "playlist"
        }
        if not url:
            deezer_obj['picture_small'] = "https://e-cdns-images.dzcdn.net/images/cover/d41d8cd98f00b204e9800998ecf8427e/56x56-000000-80-0-0.jpg"
            deezer_obj['picture_medium'] = "https://e-cdns-images.dzcdn.net/images/cover/d41d8cd98f00b204e9800998ecf8427e/250x250-000000-80-0-0.jpg"
            deezer_obj['picture_big'] = "https://e-cdns-images.dzcdn.net/images/cover/d41d8cd98f00b204e9800998ecf8427e/500x500-000000-80-0-0.jpg"
            deezer_obj['picture_xl'] = "https://e-cdns-images.dzcdn.net/images/cover/d41d8cd98f00b204e9800998ecf8427e/1000x1000-000000-80-0-0.jpg"
        return deezer_obj

    # Returns deezer song_id from spotify track_id or track dict
    def get_trackid_spotify(self, dz, track_id, fallbackSearch, spotifyTrack=None):
        if not self.spotifyEnabled:
            raise spotifyFeaturesNotEnabled
        singleTrack = False
        if not spotifyTrack:
            if (self.configFolder / 'spotifyCache.json').is_file():
                with open(self.configFolder / 'spotifyCache.json', 'r') as spotifyCache:
                    cache = json.load(spotifyCache)
            else:
                cache = {'tracks': {}, 'albums': {}}
            if str(track_id) in cache['tracks']:
                dz_track = None
                if cache['tracks'][str(track_id)]['isrc']:
                    dz_track = dz.get_track_by_ISRC(cache['tracks'][str(track_id)]['isrc'])
                    dz_id = dz_track['id'] if 'id' in dz_track and 'title' in dz_track else "0"
                    cache['tracks'][str(track_id)]['id'] = dz_id
                return (cache['tracks'][str(track_id)]['id'], dz_track, cache['tracks'][str(track_id)]['isrc'])
            singleTrack = True
            spotify_track = self.sp.track(track_id)
        else:
            spotify_track = spotifyTrack
        dz_id = "0"
        dz_track = None
        isrc = None
        if 'external_ids' in spotify_track and 'isrc' in spotify_track['external_ids']:
            try:
                dz_track = dz.get_track_by_ISRC(spotify_track['external_ids']['isrc'])
                dz_id = dz_track['id'] if 'id' in dz_track and 'title' in dz_track else "0"
                isrc = spotify_track['external_ids']['isrc']
            except:
                dz_id = dz.get_track_from_metadata(spotify_track['artists'][0]['name'], spotify_track['name'],
                                                      spotify_track['album']['name']) if fallbackSearch else "0"
        elif fallbackSearch:
            dz_id = dz.get_track_from_metadata(spotify_track['artists'][0]['name'], spotify_track['name'],
                                                  spotify_track['album']['name'])
        if singleTrack:
            cache['tracks'][str(track_id)] = {'id': dz_id, 'isrc': isrc}
            with open(self.configFolder / 'spotifyCache.json', 'w') as spotifyCache:
                json.dump(cache, spotifyCache)
        return (dz_id, dz_track, isrc)

    # Returns deezer album_id from spotify album_id
    def get_albumid_spotify(self, dz, album_id):
        if not self.spotifyEnabled:
            raise spotifyFeaturesNotEnabled
        if (self.configFolder / 'spotifyCache.json').is_file():
            with open(self.configFolder / 'spotifyCache.json', 'r') as spotifyCache:
                cache = json.load(spotifyCache)
        else:
            cache = {'tracks': {}, 'albums': {}}
        if str(album_id) in cache['albums']:
            return cache['albums'][str(album_id)]['id']
        spotify_album = self.sp.album(album_id)
        dz_album = "0"
        upc = None
        if 'external_ids' in spotify_album and 'upc' in spotify_album['external_ids']:
            try:
                dz_album = dz.get_album_by_UPC(spotify_album['external_ids']['upc'])
                dz_album = dz_album['id'] if 'id' in dz_album else "0"
                upc = spotify_album['external_ids']['upc']
            except:
                try:
                    dz_album = dz.get_album_by_UPC(int(spotify_album['external_ids']['upc']))
                    dz_album = dz_album['id'] if 'id' in dz_album else "0"
                except:
                    dz_album = "0"
        cache['albums'][str(album_id)] = {'id': dz_album, 'upc': upc}
        with open(self.configFolder / 'spotifyCache.json', 'w') as spotifyCache:
            json.dump(cache, spotifyCache)
        return dz_album


    def generate_playlist_queueitem(self, dz, playlist_id, bitrate, settings):
        if not self.spotifyEnabled:
            raise spotifyFeaturesNotEnabled
        spotify_playlist = self.sp.playlist(playlist_id)

        if len(spotify_playlist['images']):
            cover = spotify_playlist['images'][0]['url']
        else:
            cover = "https://e-cdns-images.dzcdn.net/images/cover/d41d8cd98f00b204e9800998ecf8427e/75x75-000000-80-0-0.jpg"

        playlistAPI = self._convert_playlist_structure(spotify_playlist)
        playlistAPI['various_artist'] = dz.get_artist(5080)

        extra = {}
        extra['unconverted'] = []

        tracklistTmp = spotify_playlist['tracks']['items']
        while spotify_playlist['tracks']['next']:
            spotify_playlist['tracks'] = self.sp.next(spotify_playlist['tracks'])
            tracklistTmp += spotify_playlist['tracks']['items']
        for item in tracklistTmp:
            if item['track']:
                if item['track']['explicit']:
                    playlistAPI['explicit'] = True
                extra['unconverted'].append(item['track'])

        totalSize = len(extra['unconverted'])
        if not 'explicit' in playlistAPI:
            playlistAPI['explicit'] = False
        extra['playlistAPI'] = playlistAPI
        return QIConvertable(
            playlist_id,
            bitrate,
            spotify_playlist['name'],
            spotify_playlist['owner']['display_name'],
            cover,
            playlistAPI['explicit'],
            totalSize,
            'spotify_playlist',
            settings,
            extra,
        )

    def convert_spotify_playlist(self, dz, queueItem, interface=None):
        convertPercentage = 0
        lastPercentage = 0
        if (self.configFolder / 'spotifyCache.json').is_file():
            with open(self.configFolder / 'spotifyCache.json', 'r') as spotifyCache:
                cache = json.load(spotifyCache)
        else:
            cache = {'tracks': {}, 'albums': {}}
        if interface:
            interface.send("startConversion", queueItem.uuid)
        collection = []
        for pos, track in enumerate(queueItem.extra['unconverted'], start=1):
            if queueItem.cancel:
                return
            if str(track['id']) in cache['tracks']:
                trackID = cache['tracks'][str(track['id'])]['id']
                trackAPI = None
                if cache['tracks'][str(track['id'])]['isrc']:
                    trackAPI = dz.get_track_by_ISRC(cache['tracks'][str(track['id'])]['isrc'])
            else:
                (trackID, trackAPI, isrc)  = self.get_trackid_spotify(dz, "0", queueItem.settings['fallbackSearch'], track)
                cache['tracks'][str(track['id'])] = {
                    'id': trackID,
                    'isrc': isrc
                }
            if str(trackID) == "0":
                deezerTrack = {
                    'SNG_ID': "0",
                    'SNG_TITLE': track['name'],
                    'DURATION': 0,
                    'MD5_ORIGIN': 0,
                    'MEDIA_VERSION': 0,
                    'FILESIZE': 0,
                    'ALB_TITLE': track['album']['name'],
                    'ALB_PICTURE': "",
                    'ART_ID': 0,
                    'ART_NAME': track['artists'][0]['name']
                }
            else:
                deezerTrack = dz.get_track_gw(trackID)
            deezerTrack['_EXTRA_PLAYLIST'] = queueItem.extra['playlistAPI']
            if trackAPI:
                deezerTrack['_EXTRA_TRACK'] = trackAPI
            deezerTrack['POSITION'] = pos
            deezerTrack['SIZE'] = queueItem.size
            deezerTrack['FILENAME_TEMPLATE'] = queueItem.settings['playlistTracknameTemplate']
            collection.append(deezerTrack)

            convertPercentage = (pos / queueItem.size) * 100
            if round(convertPercentage) != lastPercentage and round(convertPercentage) % 5 == 0:
                lastPercentage = round(convertPercentage)
                if interface:
                    interface.send("updateQueue", {'uuid': queueItem.uuid, 'conversion': lastPercentage})

        queueItem.extra = None
        queueItem.collection = collection

        with open(self.configFolder / 'spotifyCache.json', 'w') as spotifyCache:
            json.dump(cache, spotifyCache)

    def get_user_playlists(self, user):
        if not self.spotifyEnabled:
            raise spotifyFeaturesNotEnabled
        result = []
        playlists = self.sp.user_playlists(user)
        while playlists:
            for playlist in playlists['items']:
                result.append(self._convert_playlist_structure(playlist))
            if playlists['next']:
                playlists = self.sp.next(playlists)
            else:
                playlists = None
        return result

    def get_playlist_tracklist(self, id):
        if not self.spotifyEnabled:
            raise spotifyFeaturesNotEnabled
        playlist = self.sp.playlist(id)
        tracklist = playlist['tracks']['items']
        while playlist['tracks']['next']:
            playlist['tracks'] = self.sp.next(playlist['tracks'])
            tracklist += playlist['tracks']['items']
        playlist['tracks'] = tracklist
        return playlist


class spotifyFeaturesNotEnabled(Exception):
    pass



#!/usr/bin/env python3
import click

from deemix.app.cli import cli
from pathlib import Path

@click.command()
@click.option('--portable', is_flag=True, help='Creates the config folder in the same directory where the script is launched')
@click.option('-b', '--bitrate', default=None, help='Overwrites the default bitrate selected')
@click.option('-p', '--path', type=str, help='Downloads in the given folder')
@click.argument('url', nargs=-1, required=True)
def download(url, bitrate, portable, path):

    localpath = Path('.')
    configFolder = localpath / 'config' if portable else None
    if path is not None:
        if path == '': path = '.'
        path = Path(path)

    app = cli(path, configFolder)
    app.login()
    url = list(url)

    try:
        isfile = Path(url[0]).is_file()
    except:
        isfile = False
    if isfile:
        filename = url[0]
        with open(filename) as f:
            url = f.readlines()

    app.downloadLink(url, bitrate)
    click.echo("All done!")

if __name__ == '__main__':
    download()



import logging

from deemix.api.deezer import APIError, LyricsStatus
from deemix.utils import removeFeatures, andCommaConcat, uniqueArray, generateReplayGainString

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('deemix')

VARIOUS_ARTISTS = 5080

class Track:
    def __init__(self, dz, settings, trackAPI_gw, trackAPI=None, albumAPI_gw=None, albumAPI=None):
        self.parseEssentialData(dz, trackAPI_gw)

        self.title = trackAPI_gw['SNG_TITLE'].strip()
        if trackAPI_gw.get('VERSION') and not trackAPI_gw['VERSION'] in trackAPI_gw['SNG_TITLE']:
            self.title += " " + trackAPI_gw['VERSION'].strip()

        self.position = trackAPI_gw.get('POSITION')

        self.localTrack = int(self.id) < 0
        if self.localTrack:
            self.parseLocalTrackData(trackAPI_gw)
        else:
            self.parseData(dz, settings, trackAPI_gw, trackAPI, albumAPI_gw, albumAPI)

        # Make sure there is at least one artist
        if not 'Main' in self.artist:
            self.artist['Main'] = [self.mainArtist['name']]

        # Fix incorrect day month when detectable
        if int(self.date['month']) > 12:
            monthTemp = self.date['month']
            self.date['month'] = self.date['day']
            self.date['day'] = monthTemp
        if int(self.album['date']['month']) > 12:
            monthTemp = self.album['date']['month']
            self.album['date']['month'] = self.album['date']['day']
            self.album['date']['day'] = monthTemp

        # Add playlist data if track is in a playlist
        self.playlist = None
        if "_EXTRA_PLAYLIST" in trackAPI_gw:
            self.parsePlaylistData(trackAPI_gw["_EXTRA_PLAYLIST"], settings)

        self.singleDownload = trackAPI_gw.get('SINGLE_TRACK', False)

        self.generateMainFeatStrings()

        # Bits useful for later
        self.searched = False
        self.selectedFormat = 0
        self.dateString = None
        self.album['picUrl'] = None
        self.album['picPath'] = None
        self.album['bitrate'] = 0
        self.album['dateString'] = None
        self.artistsString = ""

    def parseEssentialData(self, dz, trackAPI_gw):
        self.id = trackAPI_gw['SNG_ID']
        self.duration = trackAPI_gw['DURATION']
        self.MD5 = trackAPI_gw['MD5_ORIGIN']
        self.mediaVersion = trackAPI_gw['MEDIA_VERSION']
        self.fallbackId = "0"
        if 'FALLBACK' in trackAPI_gw:
            self.fallbackId = trackAPI_gw['FALLBACK']['SNG_ID']
        self.filesizes = dz.get_track_filesizes(self.id)

    def parseLocalTrackData(self, trackAPI_gw):
        # Local tracks has only the trackAPI_gw page and
        # contains only the tags provided by the file
        self.album = {
            'id': "0",
            'title': trackAPI_gw['ALB_TITLE'],
            'pic': trackAPI_gw.get('ALB_PICTURE', "")
        }
        self.mainArtist = {
            'id': "0",
            'name': trackAPI_gw['ART_NAME'],
            'pic': ""
        }
        self.artists = [trackAPI_gw['ART_NAME']]
        self.artist = {
            'Main': [trackAPI_gw['ART_NAME']]
        }
        self.date = {
            'day': "00",
            'month': "00",
            'year': "XXXX"
        }
        # Defaulting all the missing data
        self.ISRC = ""
        self.album['artist'] = self.artist
        self.album['artists'] = self.artists
        self.album['barcode'] = "Unknown"
        self.album['date'] = self.date
        self.album['discTotal'] = "0"
        self.album['explicit'] = False
        self.album['genre'] = []
        self.album['label'] = "Unknown"
        self.album['mainArtist'] = self.mainArtist
        self.album['recordType'] = "album"
        self.album['trackTotal'] = "0"
        self.bpm = 0
        self.contributors = {}
        self.copyright = ""
        self.discNumber = "0"
        self.explicit = False
        self.lyrics = {}
        self.replayGain = ""
        self.trackNumber = "0"

    def parseData(self, dz, settings, trackAPI_gw, trackAPI, albumAPI_gw, albumAPI):
        self.discNumber = trackAPI_gw.get('DISK_NUMBER')
        self.explicit = bool(int(trackAPI_gw.get('EXPLICIT_LYRICS', "0")))
        self.copyright = trackAPI_gw.get('COPYRIGHT')
        self.replayGain = ""
        if 'GAIN' in trackAPI_gw:
            self.replayGain = generateReplayGainString(trackAPI_gw['GAIN'])
        self.ISRC = trackAPI_gw.get('ISRC')
        self.trackNumber = trackAPI_gw['TRACK_NUMBER']
        self.contributors = trackAPI_gw['SNG_CONTRIBUTORS']

        self.lyrics = {
            'id': int(trackAPI_gw.get('LYRICS_ID', "0")),
            'unsync': None,
            'sync': None,
            'syncID3': None
        }
        if not "LYRICS" in trackAPI_gw and self.lyrics['id'] != 0:
            logger.info(f"[{trackAPI_gw['ART_NAME']} - {self.title}] Getting lyrics")
            try:
                trackAPI_gw["LYRICS"] = dz.get_lyrics_gw(self.id)
            except APIError:
                self.lyrics['id'] = 0
        if self.lyrics['id'] != 0:
            self.lyrics['unsync'] = trackAPI_gw["LYRICS"].get("LYRICS_TEXT")
            if "LYRICS_SYNC_JSON" in trackAPI_gw["LYRICS"]:
                syncLyricsJson = trackAPI_gw["LYRICS"]["LYRICS_SYNC_JSON"]
                self.lyrics['sync'] = ""
                self.lyrics['syncID3'] = []
                timestamp = ""
                milliseconds = 0
                for line in range(len(syncLyricsJson)):
                    if syncLyricsJson[line]["line"] != "":
                        timestamp = syncLyricsJson[line]["lrc_timestamp"]
                        milliseconds = int(syncLyricsJson[line]["milliseconds"])
                    else:
                        notEmptyLine = line + 1
                        while syncLyricsJson[notEmptyLine]["line"] == "":
                            notEmptyLine = notEmptyLine + 1
                        timestamp = syncLyricsJson[notEmptyLine]["lrc_timestamp"]
                        milliseconds = int(syncLyricsJson[notEmptyLine]["milliseconds"])
                    self.lyrics['sync'] += timestamp + syncLyricsJson[line]["line"] + "\r\n"
                    self.lyrics['syncID3'].append((syncLyricsJson[line]["line"], milliseconds))

        self.mainArtist = {
            'id': trackAPI_gw['ART_ID'],
            'name': trackAPI_gw['ART_NAME'],
            'pic': trackAPI_gw.get('ART_PICTURE')
        }

        self.date = None
        if 'PHYSICAL_RELEASE_DATE' in trackAPI_gw:
            self.date = {
                'day': trackAPI_gw["PHYSICAL_RELEASE_DATE"][8:10],
                'month': trackAPI_gw["PHYSICAL_RELEASE_DATE"][5:7],
                'year': trackAPI_gw["PHYSICAL_RELEASE_DATE"][0:4]
            }

        self.album = {
            'id': trackAPI_gw['ALB_ID'],
            'title': trackAPI_gw['ALB_TITLE'],
            'pic': trackAPI_gw.get('ALB_PICTURE'),
            'barcode': "Unknown",
            'label': "Unknown",
            'explicit': False,
            'date': None,
            'genre': []
        }

        # Try the public API first (as it has more data)
        if not albumAPI:
            logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting album infos")
            try:
                albumAPI = dz.get_album(self.album['id'])
            except APIError:
                albumAPI = None

        if albumAPI:
            self.album['title'] = albumAPI['title']

            # Getting artist image ID
            # ex: https://e-cdns-images.dzcdn.net/images/artist/f2bc007e9133c946ac3c3907ddc5d2ea/56x56-000000-80-0-0.jpg
            artistPicture = albumAPI['artist']['picture_small']
            artistPicture = artistPicture[artistPicture.find('artist/') + 7:-24]
            self.album['mainArtist'] = {
                'id': albumAPI['artist']['id'],
                'name': albumAPI['artist']['name'],
                'pic': artistPicture
            }

            self.album['artist'] = {}
            self.album['artists'] = []
            for artist in albumAPI['contributors']:
                isVariousArtists = artist['id'] == VARIOUS_ARTISTS
                isMainArtist = artist['role'] == "Main"

                if not isVariousArtists or settings['albumVariousArtists'] and isVariousArtists:
                    if artist['name'] not in self.album['artists']:
                        self.album['artists'].append(artist['name'])

                    if isMainArtist or artist['name'] not in self.album['artist']['Main'] and not isMainArtist:
                        if not artist['role'] in self.album['artist']:
                            self.album['artist'][artist['role']] = []
                        self.album['artist'][artist['role']].append(artist['name'])

            if settings['removeDuplicateArtists']:
                self.album['artists'] = uniqueArray(self.album['artists'])
                for role in self.album['artist'].keys():
                    self.album['artist'][role] = uniqueArray(self.album['artist'][role])

            self.album['trackTotal'] = albumAPI['nb_tracks']
            self.album['recordType'] = albumAPI['record_type']

            self.album['barcode'] = albumAPI.get('upc', self.album['barcode'])
            self.album['label'] = albumAPI.get('label', self.album['label'])
            self.album['explicit'] = bool(albumAPI.get('explicit_lyrics', False))
            if 'release_date' in albumAPI:
                self.album['date'] = {
                    'day': albumAPI["release_date"][8:10],
                    'month': albumAPI["release_date"][5:7],
                    'year': albumAPI["release_date"][0:4]
                }
            self.album['discTotal'] = albumAPI.get('nb_disk', "1")
            self.copyright = albumAPI.get('copyright')

            if not self.album['pic']:
                # Getting album cover ID
                # ex: https://e-cdns-images.dzcdn.net/images/cover/2e018122cb56986277102d2041a592c8/56x56-000000-80-0-0.jpg
                self.album['pic'] = albumAPI['cover_small'][albumAPI['cover_small'].find('cover/') + 6:-24]

            if albumAPI.get('genres') and len(albumAPI['genres'].get('data', [])) > 0:
                for genre in albumAPI['genres']['data']:
                    self.album['genre'].append(genre['name'])
        else:
            if not albumAPI_gw:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting more album infos")
                try:
                    albumAPI_gw = dz.get_album_gw(self.album['id'])
                except APIError:
                    albumAPI_gw = None
                    raise AlbumDoesntExsists

            self.album['title'] = albumAPI_gw['ALB_TITLE']
            self.album['mainArtist'] = {
                'id': albumAPI_gw['ART_ID'],
                'name': albumAPI_gw['ART_NAME'],
                'pic': None
            }

            # albumAPI_gw doesn't contain the artist cover
            # Getting artist image ID
            # ex: https://e-cdns-images.dzcdn.net/images/artist/f2bc007e9133c946ac3c3907ddc5d2ea/56x56-000000-80-0-0.jpg
            logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting artist picture fallback")
            artistAPI = dz.get_artist(self.album['mainArtist']['id'])
            self.album['mainArtist']['pic'] = artistAPI['picture_small'][artistAPI['picture_small'].find('artist/') + 7:-24]

            self.album['artists'] = [albumAPI_gw['ART_NAME']]
            self.album['trackTotal'] = albumAPI_gw['NUMBER_TRACK']
            self.album['discTotal'] = albumAPI_gw['NUMBER_DISK']
            self.album['recordType'] = "album"
            self.album['label'] = albumAPI_gw.get('LABEL_NAME', self.album['label'])

            if albumAPI_gw.get('EXPLICIT_ALBUM_CONTENT') and albumAPI_gw['EXPLICIT_ALBUM_CONTENT'].get('EXPLICIT_LYRICS_STATUS'):
                explicitLyricsStatus = albumAPI_gw['EXPLICIT_ALBUM_CONTENT']['EXPLICIT_LYRICS_STATUS']
                self.album['explicit'] = explicitLyricsStatus in [LyricsStatus.EXPLICIT, LyricsStatus.PARTIALLY_EXPLICIT]

            if not self.album['pic']:
                self.album['pic'] = albumAPI_gw['ALB_PICTURE']
            if 'PHYSICAL_RELEASE_DATE' in albumAPI_gw:
                self.album['date'] = {
                    'day': albumAPI_gw["PHYSICAL_RELEASE_DATE"][8:10],
                    'month': albumAPI_gw["PHYSICAL_RELEASE_DATE"][5:7],
                    'year': albumAPI_gw["PHYSICAL_RELEASE_DATE"][0:4]
                }
        self.album['picType'] = 'cover'

        isAlbumArtistVariousArtists = self.album['mainArtist']['id'] == VARIOUS_ARTISTS
        self.album['mainArtist']['save'] = not isAlbumArtistVariousArtists or settings['albumVariousArtists'] and isAlbumArtistVariousArtists

        if self.album['date'] and not self.date:
            self.date = self.album['date']

        if not trackAPI:
            logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting extra track infos")
            trackAPI = dz.get_track(self.id)
        self.bpm = trackAPI['bpm']

        if not self.replayGain and 'gain' in trackAPI:
            self.replayGain = generateReplayGainString(trackAPI['gain'])
        if not self.explicit:
            self.explicit = trackAPI['explicit_lyrics']
        if not self.discNumber:
            self.discNumber = trackAPI['disk_number']

        self.artist = {}
        self.artists = []
        for artist in trackAPI['contributors']:
            isVariousArtists = artist['id'] == VARIOUS_ARTISTS
            isMainArtist = artist['role'] == "Main"

            if not isVariousArtists or len(trackAPI['contributors']) == 1 and isVariousArtists:
                if artist['name'] not in self.artists:
                    self.artists.append(artist['name'])

                if isMainArtist or artist['name'] not in self.artist['Main'] and not isMainArtist:
                    if not artist['role'] in self.artist:
                        self.artist[artist['role']] = []
                    self.artist[artist['role']].append(artist['name'])

        if settings['removeDuplicateArtists']:
            self.artists = uniqueArray(self.artists)
            for role in self.artist.keys():
                self.artist[role] = uniqueArray(self.artist[role])

        if not self.album['discTotal']:
            if not albumAPI_gw:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting more album infos")
                albumAPI_gw = dz.get_album_gw(self.album['id'])
            self.album['discTotal'] = albumAPI_gw['NUMBER_DISK']

        if not self.copyright:
            if not albumAPI_gw:
                logger.info(f"[{self.mainArtist['name']} - {self.title}] Getting more album infos")
                albumAPI_gw = dz.get_album_gw(self.album['id'])
            self.copyright = albumAPI_gw['COPYRIGHT']

    def parsePlaylistData(self, playlist, settings):
        self.playlist = {}
        if 'dzcdn.net' in playlist['picture_small']:
            url = playlist['picture_small']
            picType = url[url.find('images/')+7:]
            picType = picType[:picType.find('/')]
            self.playlist['pic'] = url[url.find(picType+'/') + len(picType)+1:-24]
            self.playlist['picType'] = picType
        else:
            self.playlist['pic'] = playlist['picture_xl']
            self.playlist['picType'] = None
        self.playlist['title'] = playlist['title']
        self.playlist['mainArtist'] = {
            'id': playlist['various_artist']['id'],
            'name': playlist['various_artist']['name'],
            'pic': playlist['various_artist']['picture_small'][
                   playlist['various_artist']['picture_small'].find('artist/') + 7:-24]
        }
        if settings['albumVariousArtists']:
            self.playlist['artist'] = {"Main": [playlist['various_artist']['name'], ]}
            self.playlist['artists'] = [playlist['various_artist']['name'], ]
        else:
            self.playlist['artist'] = {"Main": []}
            self.playlist['artists'] = []
        self.playlist['trackTotal'] = playlist['nb_tracks']
        self.playlist['recordType'] = "compile"
        self.playlist['barcode'] = ""
        self.playlist['label'] = ""
        self.playlist['explicit'] = playlist['explicit']
        self.playlist['date'] = {
            'day': playlist["creation_date"][8:10],
            'month': playlist["creation_date"][5:7],
            'year': playlist["creation_date"][0:4]
        }
        self.playlist['discTotal'] = "1"
        self.playlist['playlistId'] = playlist['id']
        self.playlist['owner'] = playlist['creator']

    # Removes featuring from the title
    def getCleanTitle(self):
        return removeFeatures(self.title)

    # Removes featuring from the album name
    def getCleanAlbumTitle(self):
        return removeFeatures(self.album['title'])

    def getFeatTitle(self):
        if self.featArtistsString and not "(feat." in self.title.lower():
            return self.title + " ({})".format(self.featArtistsString)
        return self.title

    def generateMainFeatStrings(self):
        self.mainArtistsString = andCommaConcat(self.artist['Main'])
        self.featArtistsString = ""
        if 'Featured' in self.artist:
            self.featArtistsString = "feat. "+andCommaConcat(self.artist['Featured'])

class TrackError(Exception):
    """Base class for exceptions in this module."""
    pass

class AlbumDoesntExsists(TrackError):
    pass



#!/usr/bin/env python3
import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="deemix",
    version="1.5.8",
    description="A barebone deezer downloader library",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://deemix.app",
    author="RemixDev",
    author_email="RemixDev64@gmail.com",
    license="GPL3",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["click", "pycryptodomex", "mutagen", "requests", "spotipy>=2.11.0", "eventlet"],
    entry_points={
        "console_scripts": [
            "deemix=deemix.__main__:download",
        ]
    },
)
